# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #
from time import perf_counter
from typing import Literal, Union
import logging
from functools import partial
from collections import UserDict, defaultdict

from scipy.constants import c as clight

import numpy as np
import xobjects as xo
import xtrack as xt

from .base_element import _handle_per_particle_blocks
from .beam_elements import Drift
from .general import _pkg_root
from .internal_record import new_io_buffer
from .line import Line, _is_thick, _is_collective
from .line import freeze_longitudinal as _freeze_longitudinal
from .pipeline import PipelineStatus
from .progress_indicator import progress
from .tracker_data import TrackerData

logger = logging.getLogger(__name__)

class Tracker:

    '''
    Xsuite tracker class. It is the core of the xsuite package, allows tracking
    particles in a given beam line. Methods to match particle distributions
    and to compute twiss parameters are also available.
    '''

    def __init__(
        self,
        _context=None,
        _buffer=None,
        line=None,
        compile=True,
        io_buffer=None,
        use_prebuilt_kernels=True,
        enable_pipeline_hold=False,
        track_kernel=None,
        particles_monitor_class=None,
        extra_headers=(),
        local_particle_src=None,
        _prebuilding_kernels=False,
    ):

        # Check if there are collective elements
        self.iscollective = False
        for ee in line.elements:
            if _is_collective(ee, line):
                self.iscollective = True
                break

        if enable_pipeline_hold and not self.iscollective:
            raise ValueError("`enable_pipeline_hold` is not implemented in "
                             "non-collective mode")

        if not compile and self.iscollective:
            raise NotImplementedError("Skip compilation is not implemented in "
                                      "collective mode")

        if local_particle_src is None:
            local_particle_src = xt.Particles.gen_local_particle_api()

        if not particles_monitor_class:
            particles_monitor_class = self._get_default_monitor_class()

        self.line = line
        self.particles_monitor_class = particles_monitor_class
        self.extra_headers = extra_headers
        self.local_particle_src = local_particle_src
        self._enable_pipeline_hold = enable_pipeline_hold
        self.use_prebuilt_kernels = use_prebuilt_kernels

        # Some data for collective mode prepared also for non-collective lines
        # to allow collective actions by the tracker (e.g. time-functions on knobs)
        (parts, part_names, _element_part, _element_index_in_part,
            _part_element_index, noncollective_xelements) = (
            self._split_parts_for_collective_mode(line, _buffer))

        assert len(line.element_names) == len(_element_index_in_part)
        assert len(line.element_names) == len(_element_part)
        if len(line.element_names) > 0:
            assert _element_part[-1] == len(parts) - 1
        self._parts = parts
        self._part_names = part_names
        self._element_part = _element_part
        self._element_index_in_part = _element_index_in_part

        if self.iscollective:
            # Build tracker for all non-collective elements
            # (with collective elements replaced by Drifts)
            ele_dict_non_collective = line.element_dict.copy() # need to keep the parents
            for nn, ee in zip(line.element_names, noncollective_xelements):
                ele_dict_non_collective[nn] = ee
        else:
            ele_dict_non_collective = line.element_dict

        if _prebuilding_kernels:
            element_s_locations = np.zeros(len(line.element_names))
            line_length = 0.
        else:
            element_s_locations = line.get_s_elements()
            line_length = line.get_length()

        tracker_data_base = TrackerData(
            allow_move=True, # Will move elements to the same buffer
            element_dict=ele_dict_non_collective,
            element_names=line.element_names,
            element_s_locations=element_s_locations,
            line_length=line_length,
            kernel_element_classes=None,
            extra_element_classes=(particles_monitor_class._XoStruct,),
            _context=_context,
            _buffer=_buffer,
            _no_resolve_parents=_prebuilding_kernels)
        if not _prebuilding_kernels:
            tracker_data_base._line_table = line.get_table()
            tracker_data_base._element_names_unique = tuple(
                    tracker_data_base._line_table.name[:-1]) # remove _endpoint
        line._freeze()

        if np.any([hasattr(ee, 'needs_rng') and ee.needs_rng for ee in line.elements]):
            line._needs_rng = True

        _buffer = tracker_data_base._buffer

        # Make a "marker" element to increase at_element
        if self.iscollective:
            self._zerodrift = Drift(_context=_buffer.context, length=0)

        self._track_kernel = track_kernel or {}
        self._tracker_data_cache = {}
        self._tracker_data_cache[None] = tracker_data_base

        if not _prebuilding_kernels:
            self._get_twiss_mask_markers() # to cache it

        self.line = line
        self._init_io_buffer(io_buffer)
        self.line.tracker = self

        if compile:
            _ = self.get_track_kernel_and_data_for_present_config()  # This triggers compilation

    def _init_io_buffer(self, io_buffer=None):
        if io_buffer is None:
            io_bufs = [ee.io_buffer for ee in self.line.elements
                       if getattr(ee, 'io_buffer', None) is not None]
            if len(io_bufs) == 0:
                io_buffer = new_io_buffer(_context=self._context)
            elif len(np.unique([id(buf) for buf in io_bufs])) > 1:
                raise ValueError("Different io buffers found in elements!")
            else:
                io_buffer = io_bufs[0]
        self.io_buffer = io_buffer

    def _split_parts_for_collective_mode(self, line, _buffer):

        # Split the sequence
        parts = []
        part_names = []
        _element_part = []
        _element_index_in_part=[]
        _part_element_index = defaultdict(list)
        this_part = Line(elements=[], element_names=[])
        ii_in_part = 0
        i_part = 0
        idx = 0
        for nn, ee in zip(line.element_names, line.elements):
            if not _is_collective(ee, line):
                this_part.append_element(ee, nn)
                _element_part.append(i_part)
                _element_index_in_part.append(ii_in_part)
                ii_in_part += 1
                _part_element_index[i_part].append(idx)
            else:
                if len(this_part.elements) > 0:
                    parts.append(this_part)
                    part_names.append(f'part_{i_part}_non_collective')
                    i_part += 1
                parts.append(ee)
                part_names.append(nn)
                _element_part.append(i_part)
                _element_index_in_part.append(None)
                _part_element_index[i_part].append(idx)
                i_part += 1
                this_part = Line(elements=[], element_names=[])
                ii_in_part = 0
            idx += 1
        if len(this_part.elements) > 0:
            parts.append(this_part)
            part_names.append(f'part_{i_part}_non_collective')

        # Build drifts to replace non-collective elements
        noncollective_xelements = []
        for ii, pp in enumerate(parts):
            if isinstance(pp, Line):
                noncollective_xelements += pp.elements
            else:
                if _is_thick(pp, line):
                    ldrift = pp.length
                else:
                    ldrift = 0.

                if _buffer is None:
                    if hasattr(pp, '_buffer'):
                        pp_buffer = pp._buffer
                    else:
                        pp_buffer = None
                else:
                    pp_buffer = _buffer

                noncollective_xelements.append(
                    Drift(_buffer=pp_buffer, length=ldrift))

        # Build TrackerPartNonCollective objects for non-collective parts
        for ii, pp in enumerate(parts):
            if isinstance(pp, Line):
                ele_start = _part_element_index[ii][0]
                ele_stop = _part_element_index[ii][-1] + 1
                parts[ii] = TrackerPartNonCollective(
                    tracker=self,
                    ele_start_in_tracker=ele_start,
                    ele_stop_in_tracker=ele_stop,
                )

        return (parts, part_names, _element_part, _element_index_in_part,
                _part_element_index, noncollective_xelements)

    @property
    def track_kernel(self):
        return self._track_kernel

    @property
    def _tracker_data_base(self):
        return self._tracker_data_cache[None]

    @property
    def line_element_classes(self):
        return self._tracker_data_cache[None].line_element_classes

    @property
    def config(self):
        return self.line.config

    @property
    def _buffer(self):
        return self._tracker_data_cache[None]._buffer

    @property
    def num_elements(self):
        return len(self.line.element_names)

    @property
    def _element_dict_non_collective(self):
        return self._tracker_data_cache[None]._element_dict

    @property
    def matrix_responsiveness_tol(self):
        return self.line.matrix_responsiveness_tol

    @property
    def matrix_stability_tol(self):
        return self.line.matrix_stability_tol

    @property
    def _radiation_model(self):
        return self.line._radiation_model

    @property
    def _beamstrahlung_model(self):
        return self.line._beamstrahlung_model

    @property
    def _bhabha_model(self):
        return self.line._bhabha_model

    def _invalidate(self):
        if self.iscollective:
            self._invalidated_parts = self._parts
            self._parts = None
        else:
            self._tracker_data_cache = None
        self._is_invalidated = True

    def _check_invalidated(self):
        if hasattr(self, '_is_invalidated') and self._is_invalidated:
            raise RuntimeError(
                "This tracker is not anymore valid, most probably because the corresponding line has been unfrozen. "
                "Please rebuild the tracker, for example using `line.build_tracker(...)`.")

    def _track(self, particles, *args, with_progress: Union[bool, int] = False,
               time=False, **kwargs):

        out = None

        if time:
            t0 = perf_counter()

        assert self.iscollective in (True, False)
        if (self.iscollective or self.line.enable_time_dependent_vars
            or 'log' in kwargs and kwargs['log'] is not None):
            tracking_func = self._track_with_collective
        else:
            tracking_func = self._track_no_collective

        if with_progress:
            if self.enable_pipeline_hold:
                raise ValueError("Progress indicator is not supported with pipeline hold")

            try:
                num_turns = kwargs['num_turns']
            except KeyError:
                raise ValueError('Tracking with progress indicator is only '
                                 'possible over more than one turn.')

            if with_progress is True:
                batch_size = scaling = 100
            else:
                batch_size = int(with_progress)
                scaling = with_progress if batch_size > 1 else None

            if kwargs.get('turn_by_turn_monitor') is True:
                ele_start = kwargs.get('ele_start') or 0
                ele_stop = kwargs.get('ele_stop')
                if ele_stop is None:
                    ele_stop = len(self.line)

                if ele_start >= ele_stop:
                    # we need an additional turn and space in the monitor for
                    # the incomplete turn
                    num_turns += 1
                _, monitor, _, _ = self._get_monitor(particles, True, num_turns)
                kwargs['turn_by_turn_monitor'] = monitor

            for ii in progress(
                    range(0, num_turns, batch_size),
                    desc='Tracking',
                    unit_scale=scaling,
            ):
                one_turn_kwargs = kwargs.copy()
                is_first_batch = ii == 0
                is_last_batch = ii + batch_size >= num_turns

                if is_first_batch and is_last_batch:
                    # This is the only batch, we track as normal
                    pass
                elif is_first_batch:
                    # Not the last batch, so track until the last element
                    one_turn_kwargs['ele_stop'] = None
                    one_turn_kwargs['num_turns'] = batch_size
                elif is_last_batch:
                    # Not the first batch, so track from the first element
                    one_turn_kwargs['ele_start'] = None
                    remaining_turns = num_turns % batch_size
                    if remaining_turns == 0:
                        remaining_turns = batch_size
                    one_turn_kwargs['num_turns'] = remaining_turns
                    one_turn_kwargs['_reset_log'] = False
                elif not is_first_batch and not is_last_batch:
                    # A 'middle batch', track from first to last element
                    one_turn_kwargs['num_turns'] = batch_size
                    one_turn_kwargs['ele_start'] = None
                    one_turn_kwargs['ele_stop'] = None
                    one_turn_kwargs['_reset_log'] = False

                tracking_func(particles, *args, **one_turn_kwargs)
                # particles.reorganize() # could be done in the future to optimize GPU usage
        else:
            out = tracking_func(particles, *args, **kwargs)

        if time:
            t1 = perf_counter()
            self._context.synchronize()
            self.time_last_track = t1 - t0
        else:
            self.time_last_track = None

        return out

    @property
    def particle_ref(self) -> xt.Particles:
        self._check_invalidated()
        return self.line.particle_ref

    @particle_ref.setter
    def particle_ref(self, value: xt.Particles):
        self.line.particle_ref = value

    @property
    def vars(self):
        self._check_invalidated()
        return self.line.vars

    @property
    def element_refs(self):
        self._check_invalidated()
        return self.line.element_refs

    @property
    def enable_pipeline_hold(self):
        return self._enable_pipeline_hold

    @enable_pipeline_hold.setter
    def enable_pipeline_hold(self, value):
        if not self.iscollective:
            raise ValueError(
                'enable_pipeline_hold is not supported non collective trackers')
        else:
            self._enable_pipeline_hold = value

    @property
    def _context(self):
        return self._buffer.context

    def _build_kernel(
            self,
            compile: Union[bool, Literal['force']],
            module_name=None,
            containing_dir='.',
            extra_classes=[],
            extra_kernels={},
    ):
        if compile == 'force':
            use_prebuilt_kernels = False
        elif not self._context.allow_prebuilt_kernels:  # only CPU serial
            use_prebuilt_kernels = False
        else:
            use_prebuilt_kernels = self.use_prebuilt_kernels

        if use_prebuilt_kernels:
            try:
                from xsuite import (
                    get_suitable_kernel,
                    XSK_PREBUILT_KERNELS_LOCATION,
                )
            except ImportError:
                kernel_info = None
            else:
                kernel_info = get_suitable_kernel(
                    self.config, self.line_element_classes
                )

            if kernel_info:
                module_name, modules_classes = kernel_info

                kernel_description = self.get_kernel_descriptions(
                                            modules_classes)['track_line']
                kernels = self._context.kernels_from_file(
                    module_name=module_name,
                    containing_dir=XSK_PREBUILT_KERNELS_LOCATION,
                    kernel_descriptions={'track_line': kernel_description},
                )
                return kernels['track_line']

        context = self._tracker_data_base._buffer.context

        kernel_element_classes = self._tracker_data_base.kernel_element_classes

        headers = []

        headers.extend(self.extra_headers)
        headers.append(_pkg_root.joinpath("headers/constants.h"))

        src_lines = []
        src_lines.append(
            r"""
            /*gpukern*/
            void track_line(
                /*gpuglmem*/ int8_t* buffer,
                             ElementRefData elem_ref_data,
                             ParticlesData particles,
                             int num_turns,
                             int ele_start,
                             int num_ele_track,
                             int flag_end_turn_actions,
                             int flag_reset_s_at_end_turn,
                             int flag_monitor,
                             int num_ele_line,
                             double line_length,
                /*gpuglmem*/ int8_t* buffer_tbt_monitor,
                             int64_t offset_tbt_monitor,
                /*gpuglmem*/ int8_t* io_buffer){

            #define CONTEXT_OPENMP  //only_for_context cpu_openmp
            #ifdef CONTEXT_OPENMP
                const int64_t capacity = ParticlesData_get__capacity(particles);
                const int num_threads = omp_get_max_threads();

                #ifndef XT_OMP_SKIP_REORGANIZE
                    const int64_t num_particles_to_track = ParticlesData_get__num_active_particles(particles);
                    
                    {
                        LocalParticle lpart;
                        lpart.io_buffer = io_buffer;
                        Particles_to_LocalParticle(particles, &lpart, 0, capacity);
                        check_is_active(&lpart);
                        count_reorganized_particles(&lpart);
                        LocalParticle_to_Particles(&lpart, particles, 0, capacity);
                    }
                #else // When we skip reorganize, we cannot just batch active particles
                    const int64_t num_particles_to_track = capacity;
                #endif
                
                const int64_t chunk_size = (num_particles_to_track + num_threads - 1)/num_threads; // ceil division
            #endif // CONTEXT_OPENMP
            
            #pragma omp parallel for                                                           //only_for_context cpu_openmp
            for (int chunk = 0; chunk < num_threads; chunk++) {                                //only_for_context cpu_openmp
            int64_t part_id = chunk * chunk_size;                                              //only_for_context cpu_openmp
            int64_t end_id = (chunk + 1) * chunk_size;                                         //only_for_context cpu_openmp
            if (end_id > num_particles_to_track) end_id = num_particles_to_track;              //only_for_context cpu_openmp

            int64_t part_id = 0;                                      //only_for_context cpu_serial
            int64_t part_id = blockDim.x * blockIdx.x + threadIdx.x;  //only_for_context cuda
            int64_t part_id = get_global_id(0);                       //only_for_context opencl
            int64_t end_id = 0; // unused outside of openmp           //only_for_context cpu_serial cuda opencl

            LocalParticle lpart;
            lpart.io_buffer = io_buffer;

            /*gpuglmem*/ int8_t* tbt_mon_pointer =
                            buffer_tbt_monitor + offset_tbt_monitor;
            ParticlesMonitorData tbt_monitor =
                            (ParticlesMonitorData) tbt_mon_pointer;

            int64_t part_capacity = ParticlesData_get__capacity(particles);
            if (part_id<part_capacity){
            Particles_to_LocalParticle(particles, &lpart, part_id, end_id);

            int64_t isactive = check_is_active(&lpart);

            for (int64_t iturn=0; iturn<num_turns; iturn++){

                if (!isactive){
                    break;
                }

                int64_t const ele_stop = ele_start + num_ele_track;

                #if defined(XSUITE_BACKTRACK) || defined(XSUITE_MIRROR)
                int64_t elem_idx = ele_stop - 1;
                int64_t const increm = -1;
                if (flag_end_turn_actions>0){
                    increment_at_turn_backtrack(&lpart, flag_reset_s_at_end_turn,
                                                line_length, num_ele_line);
                }
                #else
                if (flag_monitor==1){
                    ParticlesMonitor_track_local_particle(tbt_monitor, &lpart);
                }
                int64_t elem_idx = ele_start;
                int64_t const increm = 1;
                #endif

                for (; ((elem_idx >= ele_start) && (elem_idx < ele_stop)); elem_idx+=increm){
                        if (flag_monitor==2){
                            ParticlesMonitor_track_local_particle(tbt_monitor, &lpart);
                        }

                        // Get the pointer to and the type id of the `elem_idx`th
                        // element in `element_ref_data.elements`:
                        /*gpuglmem*/ void* el = ElementRefData_member_elements(elem_ref_data, elem_idx);
                        int64_t elem_type = ElementRefData_typeid_elements(elem_ref_data, elem_idx);

                        switch(elem_type){
        """
        )

        for ii, cc in enumerate(kernel_element_classes):
            ccnn = cc.__name__.replace("Data", "")
            src_lines.append(
                f"""
                        case {ii}:
"""
            )
            if ccnn == "Drift":
                src_lines.append(
                    """
                            #ifdef XTRACK_GLOBAL_XY_LIMIT
                            global_aperture_check(&lpart);
                            #endif

                            """
                )
            src_lines.append(
                f"""
                            {ccnn}_track_local_particle_with_transformations(({ccnn}Data) el, &lpart);
                            break;"""
            )

        src_lines.append(
            r"""
                        } //switch

                    // Setting the below flag will break particle losses
                    #ifndef DANGER_SKIP_ACTIVE_CHECK_AND_SWAPS

                    isactive = check_is_active(&lpart);
                    if (!isactive){
                        break;
                    }

                    #if defined(XSUITE_BACKTRACK) || defined(XSUITE_MIRROR)
                        increment_at_element(&lpart, -1);
                    #else
                        increment_at_element(&lpart, 1);
                    #endif

                    #endif //DANGER_SKIP_ACTIVE_CHECK_AND_SWAPS

                } // for elements

                if (flag_monitor==2){
                    // End of turn (element-by-element mode)
                    ParticlesMonitor_track_local_particle(tbt_monitor, &lpart);
                }

                #if defined(XSUITE_BACKTRACK) || defined(XSUITE_MIRROR)
                if (flag_monitor==1){
                    ParticlesMonitor_track_local_particle(tbt_monitor, &lpart);
                }
                # else
                if (flag_end_turn_actions>0){
                    if (isactive){
                        increment_at_turn(&lpart, flag_reset_s_at_end_turn);
                    }
                }
                #endif

            } // for turns

            LocalParticle_to_Particles(&lpart, particles, part_id, 1);

            }// if partid
            } //only_for_context cpu_openmp

            // On OpenMP we want to additionally by default reorganize all
            // the particles.
            #ifndef XT_OMP_SKIP_REORGANIZE                             //only_for_context cpu_openmp
            LocalParticle lpart;                                       //only_for_context cpu_openmp
            lpart.io_buffer = io_buffer;                               //only_for_context cpu_openmp
            Particles_to_LocalParticle(particles, &lpart, 0, capacity);//only_for_context cpu_openmp
            check_is_active(&lpart);                                   //only_for_context cpu_openmp
            count_reorganized_particles(&lpart);                       //only_for_context cpu_openmp
            LocalParticle_to_Particles(&lpart, particles, 0, capacity);//only_for_context cpu_openmp
            #endif                                                     //only_for_context cpu_openmp
        }//kernel
        """
        )

        source_track = "\n".join(src_lines)

        kernels = self.get_kernel_descriptions(kernel_element_classes)
        kernels.update(extra_kernels)

        # Compile!
        if self._context.allow_prebuilt_kernels:
            kwargs = {
                'containing_dir': containing_dir,
                'module_name': module_name,
            }
        else:
            # Saving kernels is unsupported on GPU
            kwargs = {}

        out_kernels = context.build_kernels(
            sources=[source_track],
            kernel_descriptions=kernels,
            extra_headers=self._config_to_headers() + headers,
            extra_classes=kernel_element_classes + extra_classes,
            apply_to_source=[
                partial(_handle_per_particle_blocks,
                        local_particle_src=self.local_particle_src)],
            specialize=True,
            compile=compile,
            save_source_as=f'{module_name}.c' if module_name else None,
            **kwargs,
        )
        return out_kernels['track_line']

    def get_kernel_descriptions(self, kernel_element_classes):

        tdata_type = _element_ref_data_class_from_element_classes(
            kernel_element_classes)

        kernel_descriptions = {
            "track_line": xo.Kernel(
                c_name='track_line',
                args=[
                    xo.Arg(xo.Int8, pointer=True, name="buffer"),
                    xo.Arg(tdata_type, name="tracker_data"),
                    xo.Arg(xt.Particles._XoStruct, name="particles"),
                    xo.Arg(xo.Int32, name="num_turns"),
                    xo.Arg(xo.Int32, name="ele_start"),
                    xo.Arg(xo.Int32, name="num_ele_track"),
                    xo.Arg(xo.Int32, name="flag_end_turn_actions"),
                    xo.Arg(xo.Int32, name="flag_reset_s_at_end_turn"),
                    xo.Arg(xo.Int32, name="flag_monitor"),
                    xo.Arg(xo.Int32, name='num_ele_line'),
                    xo.Arg(xo.Float64, name='line_length'),
                    xo.Arg(xo.Int8, pointer=True, name="buffer_tbt_monitor"),
                    xo.Arg(xo.Int64, name="offset_tbt_monitor"),
                    xo.Arg(xo.Int8, pointer=True, name="io_buffer"),
                ],
            )
        }

        # Random number generator init kernel
        kernel_descriptions.update(xt.Particles._kernels)

        # Multisetter
        kernel_descriptions.update(xt.MultiSetter._kernels)

        return kernel_descriptions

    def _prepare_collective_track_session(self, particles, ele_start, ele_stop,
                                       num_elements, num_turns, turn_by_turn_monitor):

        # Start position
        if particles.start_tracking_at_element >= 0:
            if ele_start != 0:
                raise ValueError("The argument ele_start is used, but particles.start_tracking_at_element is set as well. "
                                 "Please use only one of those methods.")
            ele_start = particles.start_tracking_at_element
            particles.start_tracking_at_element = -1
        if isinstance(ele_start, str):
            ele_start = self.line.element_names.index(ele_start)

        # ele_start can only have values of existing element id's,
        # but also allowed: all elements+1 (to perform end-turn actions)
        assert ele_start >= 0
        assert ele_start < self.num_elements

        # Stop position
        if num_elements is not None:
            # We are using ele_start and num_elements
            if ele_stop is not None:
                raise ValueError("Cannot use both num_elements and ele_stop!")
            if num_turns is not None:
                raise ValueError("Cannot use both num_elements and num_turns!")
            num_turns, ele_stop = np.divmod(ele_start + num_elements, self.num_elements)
            if ele_stop == 0:
                ele_stop = None
            else:
                num_turns += 1
        else:
            # We are using ele_start, ele_stop, and num_turns
            if num_turns is None:
                num_turns = 1
            else:
                assert num_turns > 0
            if isinstance(ele_stop, str):
                ele_stop = self.line.element_names.index(ele_stop)

            # If ele_stop comes before ele_start, we need to add an additional turn to
            # reach the required ele_stop
            if ele_stop == 0:
                ele_stop = None

            if ele_stop is not None and ele_stop <= ele_start:
                num_turns += 1

        if ele_stop is not None:
            assert ele_stop >= 0
            assert ele_stop <= self.num_elements

        assert num_turns >= 1

        (flag_monitor, monitor, buffer_monitor, offset_monitor
            ) = self._get_monitor(particles, turn_by_turn_monitor, num_turns)

        if particles._num_active_particles < 0:
            _context_needs_clean_active_lost_state = True
        else:
            _context_needs_clean_active_lost_state = False

        if self.line._needs_rng and not particles._has_valid_rng_state():
            particles._init_random_number_generator()

        return (ele_start, ele_stop, num_turns, flag_monitor, monitor,
                buffer_monitor, offset_monitor,
                _context_needs_clean_active_lost_state)

    def _prepare_particles_for_part(self, particles, pp,
                                    moveback_to_buffer, moveback_to_offset,
                                    _context_needs_clean_active_lost_state):
        if hasattr(self, '_slice_sets'):
            # If pyheadtail object, remove any stored slice sets
            # (they are made invalid by the xtrack elements changing zeta)
            self._slice_sets = {}

        if (hasattr(pp, 'needs_cpu') and pp.needs_cpu):
            # Move to CPU if not already there
            if (moveback_to_buffer is None
                and not isinstance(particles._buffer.context, xo.ContextCpu)):
                moveback_to_buffer = particles._buffer
                moveback_to_offset = particles._offset
                particles.move(_context=xo.ContextCpu())
                particles.reorganize()
        else:
            # Move to GPU if not already there
            if moveback_to_buffer is not None:
                particles.move(_buffer=moveback_to_buffer, _offset=moveback_to_offset)
                moveback_to_buffer = None
                moveback_to_offset = None
                if _context_needs_clean_active_lost_state:
                    particles._num_active_particles = -1
                    particles._num_lost_particles = -1

        # Hide lost particles if required by element
        _need_unhide_lost_particles = False
        if (hasattr(pp, 'needs_hidden_lost_particles')
            and pp.needs_hidden_lost_particles):
            if not particles.lost_particles_are_hidden:
                _need_unhide_lost_particles = True
            particles.hide_lost_particles()

        return _need_unhide_lost_particles, moveback_to_buffer, moveback_to_offset

    def _track_part(self, particles, pp, tt, ipp, ele_start, ele_stop, num_turns, monitor):
        ret = None
        skip = False
        stop_tracking = False

        if tt == 0 and ipp < self._element_part[ele_start]:
            # Do not track before ele_start in the first turn
            skip = True

        elif tt == 0 and self._element_part[ele_start] == ipp:
            # We are in the part that contains the start element
            i_start_in_part = self._element_index_in_part[ele_start]
            if i_start_in_part is None:
                # The start part is collective
                ret = pp.track(particles)
            else:
                # The start part is a non-collective tracker
                if (ele_stop is not None
                    and tt == num_turns - 1 and self._element_part[ele_stop] == ipp):
                    # The stop element is also in this part, so track until ele_stop
                    i_stop_in_part = self._element_index_in_part[ele_stop]
                    ret = pp.track(particles, ele_start=i_start_in_part, ele_stop=i_stop_in_part, turn_by_turn_monitor=monitor)
                    stop_tracking = True
                else:
                    # Track until end of part
                    ret = pp.track(particles, ele_start=i_start_in_part, turn_by_turn_monitor=monitor)

        elif (ele_stop is not None and ele_stop < self.num_elements
                and tt == num_turns-1 and self._element_part[ele_stop] == ipp):
            # We are in the part that contains the stop element
            i_stop_in_part = self._element_index_in_part[ele_stop]
            if i_stop_in_part is not None:
                # If not collective, track until ele_stop
                ret = pp.track(particles, num_elements=i_stop_in_part, turn_by_turn_monitor=monitor)
            stop_tracking = True

        else:
            # We are in between the part that contains the start element,
            # and the one that contains the stop element, so track normally
            if isinstance(pp, TrackerPartNonCollective):
                ret = pp.track(particles, turn_by_turn_monitor=monitor)
            else:
                ret = pp.track(particles)

        return stop_tracking, skip, ret

    def resume(self, session):
        """
        Resume a track session that had been placed on hold.
        """
        return self._track_with_collective(particles=None, _session_to_resume=session,
                                           _reset_log=False)

    def _track_with_collective(
        self,
        particles,
        ele_start=None,
        ele_stop=None,     # defaults to full lattice
        num_elements=None, # defaults to full lattice
        num_turns=None,    # defaults to 1
        turn_by_turn_monitor=None,
        freeze_longitudinal=False,
        backtrack=False,
        log=None,
        _session_to_resume=None,
        _reset_log=True,
    ):

        if ele_start is None:
            ele_start = 0

        if freeze_longitudinal:
            raise NotImplementedError('freeze_longitudinal not implemented yet'
                                      ' for collective tracking')

        if backtrack:
            raise NotImplementedError('backtrack not available for collective'
                                      ' tracking')

        if log is not None:
            if isinstance(log, (list, tuple)):
                log = Log(*log)
            elif not isinstance(log, Log):
                log = Log(log)
        if log is not None and _reset_log:
            self.line.log_last_track = {}

        self._check_invalidated()

        if (isinstance(self._buffer.context, xo.ContextCpu)
                and _session_to_resume is None):
            if not (particles._num_active_particles >= 0 and
                    particles._num_lost_particles >= 0):
                raise ValueError("Particles state is not valid to run on CPU, "
                                 "please call `particles.reorganize()` first.")

        if _session_to_resume is not None:
            if isinstance(_session_to_resume, PipelineStatus):
                _session_to_resume = _session_to_resume.data

            assert not(_session_to_resume['resumed']), (
                "This session hase been already resumed")

            assert _session_to_resume['tracker'] is self, (
                "This session was not created by this tracker")

            ele_start = _session_to_resume['ele_start']
            ele_stop = _session_to_resume['ele_stop']
            num_turns = _session_to_resume['num_turns']
            flag_monitor = _session_to_resume['flag_monitor']
            monitor = _session_to_resume['monitor']
            _context_needs_clean_active_lost_state = _session_to_resume[
                                    '_context_needs_clean_active_lost_state']
            tt_resume = _session_to_resume['tt']
            ipp_resume = _session_to_resume['ipp']
            log = _session_to_resume['log']
            _session_to_resume['resumed'] = True
        else:
            (ele_start, ele_stop, num_turns, flag_monitor, monitor,
                buffer_monitor, offset_monitor,
                _context_needs_clean_active_lost_state
                ) = self._prepare_collective_track_session(
                                particles, ele_start, ele_stop,
                                num_elements, num_turns, turn_by_turn_monitor)
            tt_resume = None
            ipp_resume = None

        for tt in range(num_turns):
            if tt_resume is not None and tt < tt_resume:
                continue

            if (flag_monitor and (ele_start == 0 or tt>0) # second condition is for delayed start
                and not _is_resume_within_turn(tt, tt_resume)):
                    monitor.track(particles)

            # Time dependent vars and energy ramping
            if self.line.enable_time_dependent_vars:
                self._handle_time_dependent_vars(particles=particles)

            if log is not None and not _is_resume_within_turn(tt, tt_resume):
                self._handle_log(_session_to_resume=_session_to_resume,
                                 particles=particles, log=log)

            moveback_to_buffer = None
            moveback_to_offset = None
            for ipp, pp in enumerate(self._parts):
                if (ipp_resume is not None and ipp < ipp_resume):
                    continue
                elif (ipp_resume is not None and ipp == ipp_resume):
                    assert particles is None
                    particles = _session_to_resume['particles']
                    pp = self._parts[ipp]
                    moveback_to_buffer = _session_to_resume['moveback_to_buffer']
                    moveback_to_offset = _session_to_resume['moveback_to_offset']
                    _context_needs_clean_active_lost_state = _session_to_resume[
                                    '_context_needs_clean_active_lost_state']
                    _need_unhide_lost_particles = _session_to_resume[
                                    '_need_unhide_lost_particles']
                    # Clear
                    tt_resume = None
                    ipp_resume = None
                else:
                    (_need_unhide_lost_particles, moveback_to_buffer,
                        moveback_to_offset) = self._prepare_particles_for_part(
                                            particles, pp,
                                            moveback_to_buffer, moveback_to_offset,
                                            _context_needs_clean_active_lost_state)

                if monitor is not None and monitor.ebe_mode == 1:
                    monitor_part = monitor
                else:
                    monitor_part = None

                # Track!
                stop_tracking, skip, returned_by_track = self._track_part(
                        particles, pp, tt, ipp, ele_start, ele_stop, num_turns, monitor_part)

                if returned_by_track is not None:
                    if returned_by_track.on_hold:

                        assert self.enable_pipeline_hold, (
                            "Hold session not enabled for this tracker.")

                        session_on_hold = {
                            'particles': particles,
                            'tracker': self,
                            'status_from_element': returned_by_track,
                            'ele_start': ele_start,
                            'ele_stop': ele_stop,
                            'num_elements': num_elements,
                            'num_turns': num_turns,
                            'flag_monitor': flag_monitor,
                            'monitor': monitor,
                            '_context_needs_clean_active_lost_state':
                                        _context_needs_clean_active_lost_state,
                            '_need_unhide_lost_particles':
                                        _need_unhide_lost_particles,
                            'moveback_to_buffer': moveback_to_buffer,
                            'moveback_to_offset': moveback_to_offset,
                            'ipp': ipp,
                            'tt': tt,
                            'resumed': False,
                            'log': log
                        }
                    return PipelineStatus(on_hold=True, data=session_on_hold)

                # Do nothing before ele_start in the first turn
                if skip:
                    continue

                # For collective parts increment at_element
                if not isinstance(pp, TrackerPartNonCollective) and not stop_tracking:
                    if moveback_to_buffer is not None: # The particles object is temporarily on CPU
                        if not hasattr(self, '_zerodrift_cpu'):
                            self._zerodrift_cpu = self._zerodrift.copy(
                                _context=particles._buffer.context)
                        self._zerodrift_cpu.track(particles, increment_at_element=True)
                    else:
                        self._zerodrift.track(particles, increment_at_element=True)

                if _need_unhide_lost_particles:
                    particles.unhide_lost_particles()

                # Break from loop over parts if stop element reached
                if stop_tracking:
                    break

            ## Break from loop over turns if stop element reached
            if stop_tracking:
                break

            if moveback_to_buffer is not None:
                particles.move(
                        _buffer=moveback_to_buffer, _offset=moveback_to_offset)
                moveback_to_buffer = None
                moveback_to_offset = None
                if _context_needs_clean_active_lost_state:
                    particles._num_active_particles = -1
                    particles._num_lost_particles = -1

            if monitor is not None and monitor.ebe_mode == 1 and (
                ele_stop is None or ele_stop == self.num_elements):
                monitor.track(particles)

            # Increment at_turn and reset at_element
            # (use the non-collective track method to perform only end-turn actions)
            self._track_no_collective(particles,
                               ele_start=self.num_elements,
                               num_elements=0)

        self.record_last_track = monitor

    def _track_no_collective(
        self,
        particles,
        ele_start=None,
        ele_stop=None,     # defaults to full lattice
        num_elements=None, # defaults to full lattice
        num_turns=None,    # defaults to 1
        turn_by_turn_monitor=None,
        freeze_longitudinal=False,
        backtrack=False,
        log=None,
        _force_no_end_turn_actions=False,
        _reset_log=True,
    ):

        self._check_invalidated()

        if log is not None:
            raise NotImplementedError(
                'log can be set only when time-dependent variables are enabled')

        if backtrack != False:
            kwargs = locals().copy()
            if isinstance(backtrack, str):
                assert backtrack == 'force'
                force_backtrack = True
            else:
                force_backtrack = False
            if not(force_backtrack) and not(self._tracker_data_base._is_backtrackable):
                raise ValueError("This line is not backtrackable.")
            kwargs.pop('self')
            kwargs.pop('backtrack')
            with xt.line._preserve_config(self):
                self.config.XSUITE_BACKTRACK = True
                return self._track_no_collective(**kwargs)

        self.local_particle_src = particles.gen_local_particle_api()

        if freeze_longitudinal:
            kwargs = locals().copy()
            kwargs.pop('self')
            kwargs.pop('freeze_longitudinal')

            with _freeze_longitudinal(self.line):
                return self._track_no_collective(**kwargs)

        if isinstance(self._buffer.context, xo.ContextCpu):
            assert (particles._num_active_particles >= 0 and
                    particles._num_lost_particles >= 0), (
                        "Particles state is not valid to run on CPU, please "
                        "call `particles.reorganize()` first."
                    )

        # Start position
        if particles.start_tracking_at_element >= 0:
            if ele_start != 0:
                raise ValueError("The argument ele_start is used, but particles.start_tracking_at_element is set as well. "
                                 "Please use only one of those methods.")
            ele_start = particles.start_tracking_at_element
            particles.start_tracking_at_element = -1
        if isinstance(ele_start, str):
            ele_start = self.line.element_names.index(ele_start)

        if ele_start is None:
            ele_start = 0

        assert ele_start >= 0
        assert ele_start <= self.num_elements

        # Logic to split the tracking turns:
        # Case 1: 0 <= start < stop <= L
        #      Track first turn from start until stop (with num_elements_first_turn=stop-start)
        # Case 2: 0 <= start < L < stop < 2L
        #      Track first turn from start until L    (with num_elements_first_turn=L-start)
        #      Track last turn from 0 until stop      (with num_elements_last_turn=stop)
        # Case 3: 0 <= start < L < stop=nL
        #      Track first turn from start until L    (with num_elements_first_turn=L-start)
        #      Track middle turns from 0 until (n-1)L (with num_middle_turns=n-1)
        # Case 4: 0 <= start < L < nL < stop
        #      Track first turn from start until L    (with num_elements_first_turn=L-start)
        #      Track middle turns from 0 until (n-1)L (with num_middle_turns=n-1)
        #      Track last turn from 0 until stop      (with num_elements_last_turn=stop)

        num_middle_turns = 0
        num_elements_last_turn = 0

        if num_elements is not None:
            # We are using ele_start and num_elements
            assert num_elements >= 0
            if ele_stop is not None:
                raise ValueError("Cannot use both num_elements and ele_stop!")
            if num_turns is not None:
                raise ValueError("Cannot use both num_elements and num_turns!")
            if num_elements + ele_start <= self.num_elements:
                # Track only the first (potentially partial) turn
                num_elements_first_turn = num_elements
            else:
                # Track the first turn until the end of the lattice
                num_elements_first_turn = self.num_elements - ele_start
                # Middle turns and potential last turn
                num_middle_turns, ele_stop = np.divmod(ele_start + num_elements, self.num_elements)
                num_elements_last_turn = ele_stop
                num_middle_turns -= 1

        else:
            # We are using ele_start, ele_stop, and num_turns
            if num_turns is None:
                num_turns = 1
            else:
                assert num_turns > 0
            if ele_stop is None:
                # Track the first turn until the end of the lattice
                # (last turn is also a full cycle, so will be treated as a middle turn)
                num_elements_first_turn = self.num_elements - ele_start
                num_middle_turns = num_turns - 1
            else:
                if isinstance(ele_stop, str):
                    ele_stop = self.line.element_names.index(ele_stop)
                assert ele_stop >= 0
                assert ele_stop <= self.num_elements
                if ele_stop <= ele_start:
                    # Correct for overflow:
                    num_turns += 1
                if num_turns == 1:
                    # Track only the first partial turn
                    num_elements_first_turn = ele_stop - ele_start
                else:
                    # Track the first turn until the end of the lattice
                    num_elements_first_turn = self.num_elements - ele_start
                    # Track the middle turns
                    num_middle_turns = num_turns - 2
                    # Track the last turn until ele_stop
                    num_elements_last_turn = ele_stop

        if self.skip_end_turn_actions or _force_no_end_turn_actions:
            flag_end_first_turn_actions = False
            flag_end_middle_turn_actions = False
        else:
            flag_end_first_turn_actions = (
                    num_elements_first_turn + ele_start == self.num_elements)
            flag_end_middle_turn_actions = True

        if num_elements_last_turn > 0:
            # One monitor record for the initial turn, num_middle_turns records for the middle turns,
            # and one for the last turn
            monitor_turns = num_middle_turns + 2
        else:
            # One monitor record for the initial turn, and num_middle_turns record for the middle turns
            monitor_turns = num_middle_turns + 1

        (flag_monitor, monitor, buffer_monitor, offset_monitor
            ) = self._get_monitor(particles, turn_by_turn_monitor, monitor_turns)

        if self.line._needs_rng and not particles._has_valid_rng_state():
            particles._init_random_number_generator()

        track_kernel, tracker_data = self.get_track_kernel_and_data_for_present_config()
        track_kernel.description.n_threads = particles._capacity

        # First turn
        assert num_elements_first_turn >= 0
        track_kernel(
            buffer=tracker_data._buffer.buffer,
            tracker_data=tracker_data._element_ref_data,
            particles=particles._xobject,
            num_turns=1,
            ele_start=ele_start,
            num_ele_track=num_elements_first_turn,
            flag_end_turn_actions=flag_end_first_turn_actions,
            flag_reset_s_at_end_turn=self.reset_s_at_end_turn,
            flag_monitor=flag_monitor,
            num_ele_line=len(tracker_data.element_names),
            line_length=tracker_data.line_length,
            buffer_tbt_monitor=buffer_monitor,
            offset_tbt_monitor=offset_monitor,
            io_buffer=self.io_buffer.buffer,
        )

        # Middle turns
        if num_middle_turns > 0:
            assert self.num_elements > 0
            track_kernel(
                buffer=tracker_data._buffer.buffer,
                tracker_data=tracker_data._element_ref_data,
                particles=particles._xobject,
                num_turns=num_middle_turns,
                ele_start=0, # always full turn
                num_ele_track=self.num_elements, # always full turn
                flag_end_turn_actions=flag_end_middle_turn_actions,
                flag_reset_s_at_end_turn=self.reset_s_at_end_turn,
                flag_monitor=flag_monitor,
                num_ele_line=len(tracker_data.element_names),
                line_length=tracker_data.line_length,
                buffer_tbt_monitor=buffer_monitor,
                offset_tbt_monitor=offset_monitor,
                io_buffer=self.io_buffer.buffer,
            )

        # Last turn, only if incomplete
        if num_elements_last_turn > 0:
            assert num_elements_last_turn > 0
            track_kernel(
                buffer=tracker_data._buffer.buffer,
                tracker_data=tracker_data._element_ref_data,
                particles=particles._xobject,
                num_turns=1,
                ele_start=0,
                num_ele_track=num_elements_last_turn,
                flag_end_turn_actions=False,
                flag_reset_s_at_end_turn=self.reset_s_at_end_turn,
                flag_monitor=flag_monitor,
                num_ele_line=len(tracker_data.element_names),
                line_length=tracker_data.line_length,
                buffer_tbt_monitor=buffer_monitor,
                offset_tbt_monitor=offset_monitor,
                io_buffer=self.io_buffer.buffer,
            )

        self.record_last_track = monitor

    @staticmethod
    def _get_default_monitor_class():
        return xt.ParticlesMonitor

    def _get_monitor(self, particles, turn_by_turn_monitor, num_turns):

        if turn_by_turn_monitor is None or turn_by_turn_monitor is False:
            flag_monitor = 0
            monitor = None
            buffer_monitor = particles._buffer.buffer  # I just need a valid buffer
            offset_monitor = 0
        elif turn_by_turn_monitor is True:
            flag_monitor = 1
            # TODO Assumes at_turn starts from zero, to be generalized
            monitor = self.particles_monitor_class(
                _context=particles._buffer.context,
                start_at_turn=0,
                stop_at_turn=num_turns,
                particle_id_range=particles.get_active_particle_id_range()
            )
            buffer_monitor = monitor._buffer.buffer
            offset_monitor = monitor._offset
        elif turn_by_turn_monitor == 'ONE_TURN_EBE':
            (_, monitor, buffer_monitor, offset_monitor
                ) = self._get_monitor(particles, turn_by_turn_monitor=True,
                                      num_turns=len(self.line.element_names)+1)
            monitor.ebe_mode = 1
            flag_monitor = 2
        elif isinstance(turn_by_turn_monitor, self.particles_monitor_class):
            if turn_by_turn_monitor.ebe_mode == 1:
                flag_monitor = 2
            else:
                flag_monitor = 1
            monitor = turn_by_turn_monitor
            buffer_monitor = monitor._buffer.buffer
            offset_monitor = monitor._offset
        else:
            raise ValueError('Please provide a valid monitor object')

        return flag_monitor, monitor, buffer_monitor, offset_monitor

    def to_binary_file(self, path):

       raise NotImplementedError('to_binary_file not implemented anymore')

    @classmethod
    def from_binary_file(cls, path, particles_monitor_class=None, **kwargs) -> 'Tracker':
        raise NotImplementedError('from_binary_file not implemented anymore')

    def _hashable_config(self):
        items = ((k, v) for k, v in self.config.items() if v is not False)
        return tuple(sorted(items))

    def _config_to_headers(self):
        headers = []
        for k, v in self.config.items():
            if not isinstance(v, bool):
                headers.append(f'#define {k} {v}')
            elif v is True:
                headers.append(f'#define {k}')
            else:
                headers.append(f'#undef {k}')
        return headers

    def _get_twiss_mask_markers(self):
        if hasattr(self._tracker_data_base, 'mask_markers_for_twiss'):
            return self._tracker_data_base.mask_markers_for_twiss
        tt = self.line.get_table()
        mask_twiss = np.ones(len(tt) + 1, dtype=bool)
        if len(tt) == 0: return mask_twiss
        mask_twiss[:-1] = tt.element_type == 'Marker'
        self._tracker_data_base.mask_markers_for_twiss = mask_twiss
        return mask_twiss

    def get_track_kernel_and_data_for_present_config(self):

        hash_config = self._hashable_config()

        if hash_config not in self.track_kernel:
            new_kernel = self._build_kernel(compile=True)
            self.track_kernel[hash_config] = new_kernel

        out_kernel = self.track_kernel[hash_config]

        if hash_config not in self._tracker_data_cache:
            kernel_element_classes = _element_classes_from_track_kernel(out_kernel)
            td_base = self._tracker_data_base
            td = TrackerData(
                element_dict=td_base._element_dict,
                element_names=td_base._element_names,
                element_s_locations=td_base.element_s_locations,
                line_length=td_base.line_length,
                cache=td_base.cache.copy(),
                kernel_element_classes=kernel_element_classes,
                extra_element_classes=td_base.extra_element_classes,
                _context=self._context,
                _buffer=self._buffer)

            self._tracker_data_cache[hash_config] = td

        out_tracker_data = self._tracker_data_cache[hash_config]

        # sanity check
        assert (len(_element_classes_from_track_kernel(out_kernel))
                == len(out_tracker_data.kernel_element_classes))

        return out_kernel, out_tracker_data

    @property
    def reset_s_at_end_turn(self):
        return self.line.reset_s_at_end_turn

    @reset_s_at_end_turn.setter
    def reset_s_at_end_turn(self, value):
        self.line.reset_s_at_end_turn = value

    @property
    def skip_end_turn_actions(self):
        return self.line.skip_end_turn_actions

    @skip_end_turn_actions.setter
    def skip_end_turn_actions(self, value):
        self.line.skip_end_turn_actions = value

    def __dir__(self):
        return list(set(object.__dir__(self) + dir(self.line)))

    def __getstate__(self):
        if not isinstance(self._context, xo.ContextCpu):
            raise TypeError("Only CPU trackers can be pickled.")

        # Remove the compiled kernels from the state
        state = self.__dict__.copy()
        state['_track_kernel'].clear()
        return state

    def check_compatibility_with_prebuilt_kernels(self):
        from xsuite import get_suitable_kernel
        get_suitable_kernel(
            config=self.line.config,
            line_element_classes=self.line_element_classes,
            verbose=True)

    def _handle_time_dependent_vars(self, particles):

        # Find first active particle
        state = particles.state
        if isinstance(particles._context, xo.ContextPyopencl):
            state = state.get()
        ii_first_active = int((state > 0).argmax())
        if ii_first_active == 0 and particles._xobject.state[0] <= 0:
            # No active particles
            at_turn = 0 # convenient for multi-turn injection
        else:
            at_turn = particles._xobject.at_turn[ii_first_active]

        if self.line.energy_program is not None:
            t_turn = self.line.energy_program.get_t_s_at_turn(at_turn)
        else:
            beta0 = particles._xobject.beta0[ii_first_active]
            t_turn = (at_turn * self._tracker_data_base.line_length
                    / (beta0 * clight))

        # Clean leftover from previous trackings
        if (self.line._t_last_update_time_dependent_vars and
            t_turn < self.line._t_last_update_time_dependent_vars):
            self.line._t_last_update_time_dependent_vars = None

        if (self.line._t_last_update_time_dependent_vars is None
            or self.line.dt_update_time_dependent_vars is None
            or t_turn > self.line._t_last_update_time_dependent_vars
                        + self.line.dt_update_time_dependent_vars):
            self.line._t_last_update_time_dependent_vars = t_turn
            self.vars['t_turn_s'] = t_turn

            if self.line.energy_program is not None:
                p0c = self.line.particle_ref._xobject.p0c[0]
                particles.update_p0c_and_energy_deviations(p0c, update_pxpy=True)

    def _handle_log(self, _session_to_resume, particles, log):
        if _session_to_resume is not None:
            plog = _session_to_resume['particles']
        else:
            plog = particles
        for kk in log:
            if log[kk] == None:
                if kk not in self.line.log_last_track:
                    self.line.log_last_track[kk] = []
                self.line.log_last_track[kk].append(self.line.vv[kk])
            else:
                ff = log[kk]
                val = ff(self.line, plog)
                if hasattr(ff, '_store'):
                    for nn in ff._store:
                        if nn not in self.line.log_last_track:
                            self.line.log_last_track[nn] = []
                        self.line.log_last_track[nn].append(val[nn])
                else:
                    if kk not in self.line.log_last_track:
                        self.line.log_last_track[kk] = []
                    self.line.log_last_track[kk].append(val)


class TrackerConfig(UserDict):

    def __setitem__(self, idx, val):
        if val is False and idx in self:
            del(self[idx]) # We don't want to store flags that are False
        else:
            super(TrackerConfig, self).__setitem__(idx, val)

    def __setattr__(self, idx, val):
        if idx == 'data':
            object.__setattr__(self, idx, val)
        elif val is not False and val is not None:
            self.data[idx] = val
        elif idx in self:
            del(self.data[idx])

    def __getattr__(self, idx):
        if idx == 'data':
            return object.__getattribute__(self, idx)
        if idx in self.data:
            return self.data[idx]
        else:
            raise AttributeError(f'No attribute {idx}')

    def update(self, other, **kwargs):
        super().update(other, **kwargs)
        keys_for_none_vals = [k for k, v in self.items() if v is False]
        for k in keys_for_none_vals:
            del self[k]


class TrackerPartNonCollective:
    def __init__(self, tracker, ele_start_in_tracker, ele_stop_in_tracker):
        self.tracker = tracker
        self.ele_start_in_tracker = ele_start_in_tracker
        self.ele_stop_in_tracker = ele_stop_in_tracker

    def track(self, particles, ele_start=None, ele_stop=None, num_elements=None, turn_by_turn_monitor=None):
        if ele_start is None:
            temp_ele_start = self.ele_start_in_tracker
        else:
            temp_ele_start = self.ele_start_in_tracker + ele_start

        if ele_stop is None:
            temp_ele_stop = self.ele_stop_in_tracker
        else:
            temp_ele_stop = self.ele_start_in_tracker + ele_stop

        if num_elements is None:
            temp_num_elements = temp_ele_stop - temp_ele_start
        else:
            temp_num_elements = num_elements

        self.tracker._track_no_collective(particles, ele_start=temp_ele_start,
                           num_elements=temp_num_elements, turn_by_turn_monitor=turn_by_turn_monitor,
                           _force_no_end_turn_actions=True)

    def __repr__(self):
        return (f'TrackerPartNonCollective({self.ele_start_in_tracker}, '
                f'{self.ele_stop_in_tracker})')


def _element_classes_from_track_kernel(kernel):
    assert kernel.description.args[1].name == 'tracker_data'
    kernel_tracker_data_type = kernel.description.args[1].atype
    kernel_element_ref_class = kernel_tracker_data_type.elements.ftype._itemtype
    kernel_element_classes = kernel_element_ref_class._reftypes
    return kernel_element_classes

def _element_ref_data_class_from_element_classes(element_classes):

    # extract XoStruct if needed
    element_classes_xostruct = []
    for cc in element_classes:
        if issubclass(cc, xo.Struct):
            element_classes_xostruct.append(cc)
        else:
            element_classes_xostruct.append(cc._XoStruct)

    # We sort the element classes by name for consistency
    # This is needed for reproducibility in Xboinc
    element_classes_xostruct = sorted(element_classes_xostruct, key=lambda s: s._DressingClass.__name__)

    class ElementRefClass(xo.UnionRef):
        _reftypes = element_classes_xostruct

    class ElementRefData(xo.Struct):
            elements = ElementRefClass[:]
            names = xo.String[:]

    return ElementRefData


class Log(dict):

    def __init__(self, *args, **kwargs):
        self.__dict__ = self
        self.update(kwargs)
        unnamed_indx = 0
        for arg in args:
            if isinstance(arg, str):
                self[arg] = None
            else:
                self[f'_unnamed_{unnamed_indx}'] = arg
                unnamed_indx += 1

def _is_resume_within_turn(tt, tt_resume):
    return tt_resume is not None and tt == tt_resume


import xtrack as xt
import numpy as np

env = xt.Environment()
env.particle_ref = xt.Particles(p0c=2e9)

n_bends_per_cell = 6
n_cells_par_arc = 3
n_arcs = 3

n_bends = n_bends_per_cell * n_cells_par_arc * n_arcs

env.vars({
    'l.mq': 0.5,
    'kqf': 0.027,
    'kqd': -0.0271,
    'l.mb': 10,
    'l.ms': 0.3,
    'k2sf': 0.001,
    'k2sd': -0.001,
    'angle.mb': 2 * np.pi / n_bends,
    'k0.mb': 'angle.mb / l.mb',
    'k0l.corrector': 0,
    'k1sl.corrector': 0,
    'l.halfcell': 38,
})

env.new('mb', xt.Bend, length='l.mb', k0='k0.mb', h='k0.mb')
env.new('mq', xt.Quadrupole, length='l.mq')
env.new('ms', xt.Sextupole, length='l.ms')
env.new('corrector', xt.Multipole, knl=[0], length=0.1)

girder = env.new_line(components=[
    env.place('mq', at=1),
    env.place('ms', at=0.8, from_='mq'),
    env.place('corrector', at=-0.8, from_='mq'),
])

girder_f = girder.clone(name='f')
girder_d = girder.clone(name='d', mirror=True)
env.set('mq.f', k1='kqf')
env.set('mq.d', k1='kqd')

halfcell = env.new_line(components=[

    # End of the half cell (will be mid of the cell)
    env.new('mid', xt.Marker, at='l.halfcell'),

    # Bends
    env.new('mb.2', 'mb', at='l.halfcell / 2'),
    env.new('mb.1', 'mb', at='-l.mb - 1', from_='mb.2'),
    env.new('mb.3', 'mb', at='l.mb + 1', from_='mb.2'),

    # Quadrupoles, sextupoles and correctors
    env.place(girder_d, at=1.2),
    env.place(girder_f, at='l.halfcell - 1.2'),

])


hcell_left = halfcell.replicate(name='l', mirror=True)
hcell_right = halfcell.replicate(name='r')

cell = env.new_line(components=[
    env.new('start', xt.Marker),
    hcell_left,
    hcell_right,
    env.new('end', xt.Marker),
])

opt = cell.match(
    method='4d',
    vary=xt.VaryList(['kqf', 'kqd'], step=1e-5),
    targets=xt.TargetSet(
        qx=0.333333,
        qy=0.333333,
    ))
tw_cell = cell.twiss4d()


env.vars({
    'kqf.ss': 0.027 / 2,
    'kqd.ss': -0.0271 / 2,
})

halfcell_ss = env.new_line(components=[

    env.new('mid.ss', xt.Marker, at='l.halfcell'),

    env.new('mq.ss.d', 'mq', k1='kqd.ss', at = '0.5 + l.mq / 2'),
    env.new('mq.ss.f', 'mq', k1='kqf.ss', at = 'l.halfcell - l.mq / 2 - 0.5'),

    env.new('corrector.ss.v', 'corrector', at=0.75, from_='mq.ss.d'),
    env.new('corrector.ss.h', 'corrector', at=-0.75, from_='mq.ss.f')
])

hcell_left_ss = halfcell_ss.replicate(name='l', mirror=True)
hcell_right_ss = halfcell_ss.replicate(name='r')
cell_ss = env.new_line(components=[
    env.new('start.ss', xt.Marker),
    hcell_left_ss,
    hcell_right_ss,
    env.new('end.ss', xt.Marker),
])

opt = cell_ss.match(
    solve=False,
    method='4d',
    vary=xt.VaryList(['kqf.ss', 'kqd.ss'], step=1e-5),
    targets=xt.TargetSet(
        betx=tw_cell.betx[-1], bety=tw_cell.bety[-1], at='start.ss',
    ))
opt.solve()


arc = env.new_line(components=[
    cell.replicate(name='cell.1'),
    cell.replicate(name='cell.2'),
    cell.replicate(name='cell.3'),
])


ss = env.new_line(components=[
    cell_ss.replicate('cell.1'),
    cell_ss.replicate('cell.2'),
])

ring = env.new_line(components=[
    arc.replicate(name='arc.1'),
    ss.replicate(name='ss.1'),
    arc.replicate(name='arc.2'),
    ss.replicate(name='ss.2'),
    arc.replicate(name='arc.3'),
    ss.replicate(name='ss.3'),
])

## Insertion

env.vars({
    'k1.q1': 0.025,
    'k1.q2': -0.025,
    'k1.q3': 0.025,
    'k1.q4': -0.02,
    'k1.q5': 0.025,
})

half_insertion = env.new_line(components=[

    # Start-end markers
    env.new('ip', xt.Marker),
    env.new('e.insertion', xt.Marker, at=76),

    # Quads
    env.new('mq.1', xt.Quadrupole, k1='k1.q1', length='l.mq', at = 20),
    env.new('mq.2', xt.Quadrupole, k1='k1.q2', length='l.mq', at = 25),
    env.new('mq.3', xt.Quadrupole, k1='k1.q3', length='l.mq', at=37),
    env.new('mq.4', xt.Quadrupole, k1='k1.q4', length='l.mq', at=55),
    env.new('mq.5', xt.Quadrupole, k1='k1.q5', length='l.mq', at=73),

    # Dipole correctors (will use h and v on the same corrector)
    env.new('corrector.ss.1', 'corrector', at=0.75, from_='mq.1'),
    env.new('corrector.ss.2', 'corrector', at=-0.75, from_='mq.2'),
    env.new('corrector.ss.3', 'corrector', at=0.75, from_='mq.3'),
    env.new('corrector.ss.4', 'corrector', at=-0.75, from_='mq.4'),
    env.new('corrector.ss.5', 'corrector', at=0.75, from_='mq.5'),

])

tw_arc = arc.twiss4d()

opt = half_insertion.match(
    solve=False,
    betx=tw_arc.betx[0], bety=tw_arc.bety[0],
    alfx=tw_arc.alfx[0], alfy=tw_arc.alfy[0],
    init_at='e.insertion',
    start='ip', end='e.insertion',
    vary=xt.VaryList(['k1.q1', 'k1.q2', 'k1.q3', 'k1.q4'], step=1e-5),
    targets=[
        xt.TargetSet(alfx=0, alfy=0, at='ip'),
        xt.Target(lambda tw: tw.betx[0] - tw.bety[0], 0),
        xt.Target(lambda tw: tw.betx.max(), xt.LessThan(400)),
        xt.Target(lambda tw: tw.bety.max(), xt.LessThan(400)),
        xt.Target(lambda tw: tw.betx.min(), xt.GreaterThan(2)),
        xt.Target(lambda tw: tw.bety.min(), xt.GreaterThan(2)),
    ]
)
opt.step(40)
opt.solve()

insertion = env.new_line([
    half_insertion.replicate('l', mirror=True),
    half_insertion.replicate('r')])



ring2 = env.new_line(components=[
    arc.replicate(name='arcc.1'),
    ss.replicate(name='sss.2'),
    arc.replicate(name='arcc.2'),
    insertion,
    arc.replicate(name='arcc.3'),
    ss.replicate(name='sss.3')
])


# # Check buffer behavior
ring2_sliced = ring2.select()
ring2_sliced.cut_at_s(np.arange(0, ring2.get_length(), 0.5))


import matplotlib.pyplot as plt
plt.close('all')
for ii, rr in enumerate([ring, ring2_sliced]):

    tw = rr.twiss4d()

    fig = plt.figure(ii, figsize=(6.4*1.2, 4.8))
    ax1 = fig.add_subplot(2, 1, 1)
    pltbet = tw.plot('betx bety', ax=ax1)
    ax2 = fig.add_subplot(2, 1, 2, sharex=ax1)
    pltdx = tw.plot('dx', ax=ax2)
    fig.subplots_adjust(right=.85)
    pltbet.move_legend(1.2,1)
    pltdx.move_legend(1.2,1)

ring2.survey().plot()


plt.show()






#!/usr/bin/env python
# coding: utf-8

# In[1]:


get_ipython().run_line_magic('matplotlib', 'widget')


# In[2]:


import xtrack as xt
import numpy as np

import matplotlib.pyplot as plt


# In[3]:


env = xt.Environment()
env.particle_ref = xt.Particles(p0c=2e9)


# ### Bumber of bends

# In[4]:


n_bends_per_cell = 6
n_cells_par_arc = 3
n_arcs = 3

n_bends = n_bends_per_cell * n_cells_par_arc * n_arcs


# ## Arc cell

# In[5]:


# Parameters
env.vars({
    'l.mq': 0.5,
    'kqf': 0.027,
    'kqd': -0.0271,
    'l.mb': 10,
    'l.ms': 0.3,
    'k2sf': 0.001,
    'k2sd': -0.001,
    'angle.mb': 2 * np.pi / n_bends,
    'k0.mb': 'angle.mb / l.mb',
    'k0l.corrector': 0,
    'k1sl.corrector': 0,
    'l.halfcell': 38,
})


# In[6]:


# Basic magnet types
env.new('mb', xt.Bend, length='l.mb', k0='k0.mb', h='k0.mb')
env.new('mq', xt.Quadrupole, length='l.mq')
env.new('ms', xt.Sextupole, length='l.ms')
env.new('corrector', xt.Multipole, knl=[0], ksl=[0])

# Arc quadrupoles
env.new('mq.f', 'mq', k1='kqf')
env.new('mq.d', 'mq', k1='kqd')

# Arc sextupoles
env.new('ms.d', 'ms', k2='k2sf')
env.new('ms.f', 'ms', k2='k2sd');


# In[7]:


halfcell = env.new_line(components=[
    # End of the half cell (will be mid of the cell)
    env.new('mid', xt.Marker, at='l.halfcell'),
    # Bends
    env.new('mb.2', 'mb', at='l.halfcell / 2'),
    env.new('mb.1', 'mb', at='-l.mb - 1', from_='mb.2'),
    env.new('mb.3', 'mb', at='l.mb + 1', from_='mb.2'),
    # Quads
    env.place('mq.d', at = '0.5 + l.mq / 2'),
    env.place('mq.f', at = 'l.halfcell - l.mq / 2 - 0.5'),
    # Sextupoles
    env.new('ms.d', 'ms', k2='k2sf', at=1.2, from_='mq.d'),
    env.new('ms.f', 'ms', k2='k2sd', at=-1.2, from_='mq.f'),
    # Dipole correctors
    env.new('corrector.v', 'corrector', at=0.75, from_='mq.d'),
    env.new('corrector.h', 'corrector', at=-0.75, from_='mq.f')
])


# In[8]:


halfcell.survey().plot()


# In[10]:


# Build full fodo cell

cell = env.new_line(components=[
    env.new('start', xt.Marker),
    -halfcell,
    halfcell,
    env.new('end', xt.Marker),
])
cell.survey().plot()


# In[11]:


# Twiss cell
cell.twiss4d().plot()


# In[12]:


# March phase advance
opt = cell.match(
    solve=False,
    method='4d',
    vary=xt.VaryList(['kqf', 'kqd'], step=1e-5),
    targets=xt.TargetSet(
        qx=0.333333,
        qy=0.333333,
    ))

print('Before match:')
opt.target_status()

opt.solve()

print('After match:')
opt.target_status()

print('\nMatch history')
opt.log()


# In[13]:


# Twiss after match
tw_cell = cell.twiss4d()


# In[14]:


tw_cell.plot()


# In[15]:


tw_cell.plot('mux muy')


# ## Straight section cell

# In[16]:


env.vars({
    'kqf.ss': 0.027 / 2,
    'kqd.ss': -0.0271 / 2,
})

halfcell_ss = env.new_line(components=[
    env.new('mid', xt.Marker, at='l.halfcell'),
    env.new('mq.ss.d', 'mq', k1='kqd.ss', at = '0.5 + l.mq / 2'),
    env.new('mq.ss.f', 'mq', k1='kqf.ss', at = 'l.halfcell - l.mq / 2 - 0.5'),
    env.new('corrector.ss.v', 'corrector', at=0.75, from_='mq.ss.d'),
    env.new('corrector.ss.h', 'corrector', at=-0.75, from_='mq.ss.f')
])

hcell_left_ss = halfcell_ss.replicate(name='l', mirror=True)
hcell_right_ss = halfcell_ss.replicate(name='r')
cell_ss = env.new_line(components=[
    env.new('start.ss', xt.Marker),
    hcell_left_ss,
    hcell_right_ss,
    env.new('end.ss', xt.Marker),
])

cell_ss.survey().plot()


# In[17]:


# March to the same betas as the edge of the arc cell
opt = cell_ss.match(
    method='4d',
    vary=xt.VaryList(['kqf.ss', 'kqd.ss'], step=1e-5),
    targets=xt.TargetSet(
        betx=tw_cell.betx[-1], bety=tw_cell.bety[-1], at='start.ss',
    ))


# ## Assemble a ring

# In[18]:


arc = 3 * cell
arc.survey().plot()


# In[19]:


straight_section = 2*cell_ss
straight_section.survey().plot()


# In[20]:


ring = 3 * (arc + straight_section)

ring.survey().plot()


# # Inspect optics

# In[21]:


tw = ring.twiss4d()
fig = plt.figure(figsize=(6.4*1.2, 4.8))
ax1 = fig.add_subplot(2, 1, 1)
pltbet = tw.plot('betx bety', ax=ax1)

ax2 = fig.add_subplot(2, 1, 2, sharex=ax1)
pltdx = tw.plot('dx', ax=ax2)

fig.subplots_adjust(right=.85)
pltbet.move_legend(1.2,1)
pltdx.move_legend(1.2,1)


# ## Build and insertion

# We want 40 m of free space with round betas to install an experiment

# In[22]:


env.vars({
    'k1.q1': 0.025,
    'k1.q2': -0.025,
    'k1.q3': 0.025,
    'k1.q4': -0.02,
    'k1.q5': 0.025,
})
env.new('mq.1', 'mq', k1='k1.q1')

half_insertion = env.new_line(components=[

    # Start-end markers
    env.new('ip', xt.Marker),
    env.new('e.insertion', xt.Marker, at=76), # Same insertion length as initial straight

    # Quads
    env.new('mq.1', 'mq', k1='k1.q1', at = 20),
    env.new('mq.2', 'mq', k1='k1.q2', at = 25),
    env.new('mq.3', 'mq', k1='k1.q3', at=37),
    env.new('mq.4', 'mq', k1='k1.q4', at=55),
    env.new('mq.5', 'mq', k1='k1.q5', at=73),

    # Dipole correctors (will use h and v on the same corrector)
    env.new('corrector.ss.1', 'corrector', at=0.75, from_='mq.1'),
    env.new('corrector.ss.2', 'corrector', at=-0.75, from_='mq.2'),
    env.new('corrector.ss.3', 'corrector', at=0.75, from_='mq.3'),
    env.new('corrector.ss.4', 'corrector', at=-0.75, from_='mq.4'),
    env.new('corrector.ss.5', 'corrector', at=0.75, from_='mq.5'),

])


# In[23]:


# Match the optics of the insertion
tw_arc = arc.twiss4d()

opt = half_insertion.match(
    solve=False,
    betx=tw_arc.betx[0], bety=tw_arc.bety[0],
    alfx=tw_arc.alfx[0], alfy=tw_arc.alfy[0],
    init_at='e.insertion',
    start='ip', end='e.insertion',
    vary=xt.VaryList(['k1.q1', 'k1.q2', 'k1.q3', 'k1.q4', 'k1.q5'], step=1e-5),
    targets=[
        xt.TargetSet(alfx=0, alfy=0, at='ip'),
        xt.Target(lambda tw: tw['betx', 'ip'] - tw['bety', 'ip'], 0),
        xt.Target(lambda tw: tw.betx.max(), xt.LessThan(400)),
        xt.Target(lambda tw: tw.bety.max(), xt.LessThan(400)),
        xt.Target(lambda tw: tw.betx.min(), xt.GreaterThan(2)),
        xt.Target(lambda tw: tw.bety.min(), xt.GreaterThan(2)),
    ]
)
opt.step(40)
opt.solve()


# In[26]:


insertion = -half_insertion + half_insertion

insertion.twiss4d().plot()


# ## Build a ring wih the insertion

# In[27]:


ring_2 = 2 * (arc + straight_section) + arc + insertion


# In[28]:


ring_2.survey().plot()


# In[29]:


tw = ring_2.twiss4d()
fig = plt.figure(figsize=(6.4*1.2, 4.8))
ax1 = fig.add_subplot(2, 1, 1)
pltbet = tw.plot('betx bety', ax=ax1)

ax2 = fig.add_subplot(2, 1, 2, sharex=ax1)
pltdx = tw.plot('dx', ax=ax2)

fig.subplots_adjust(right=.85)
pltbet.move_legend(1.2,1)
pltdx.move_legend(1.2,1)


# In[39]:


# Make a shallow copy a and slice to get optics every 0.5 m
ring_2_sliced = ring_2.select()
# ring_2_sliced.replace_all_repeated_elements()


# In[36]:


ring_2_sliced.discard_tracker()


# In[38]:


ring_2_sliced.element_names


# In[ ]:


ring_2_sliced.cut_at_s(np.arange(0, ring_2.get_length(), 0.5))


# In[ ]:


tw = ring_2_sliced.twiss4d()
fig = plt.figure(figsize=(6.4*1.2, 4.8))
ax1 = fig.add_subplot(2, 1, 1)
pltbet = tw.plot('betx bety', ax=ax1)

ax2 = fig.add_subplot(2, 1, 2, sharex=ax1)
pltdx = tw.plot('dx', ax=ax2)

fig.subplots_adjust(right=.85)
pltbet.move_legend(1.2,1)
pltdx.move_legend(1.2,1)


# In[ ]:


1/np.sqrt(tw.momentum_compaction_factor)


# In[ ]:


tw.qx


# In[ ]:






import xtrack as xt

m = xt.Multipole(knl=[1,2, 3], hxl=0.1, hyl=0.2, length=0.3)
d = xt.Drift(length=1.0)

pref1 = xt.Particles(mass0=3.0, beta0=0.1, q0=2.)
pref2 = xt.Particles(mass0=2.0, beta0=0.2, q0=5.)

p1 = pref1.copy()
p2 = pref2.copy()

p1.delta = 0.1
p1.x = 0.1
p1.y = 0.2



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib
import json
import pickle
import numpy as np

import xtrack as xt
import xobjects as xo

import ducktrack as dtk

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../../test_data').absolute()

fname_line_particles = test_data_folder.joinpath('lhc_no_bb/line_and_particle.json')
rtol_100turns = 1e-9; atol_100turns=5e-11

# fname_line_particles = test_data_folder.joinpath(
#                                   './lhc_with_bb/line_and_particle.json')
# rtol_100turns = 1e-9; atol_100turns=9e-10

#fname_line_particles = test_data_folder.joinpath(
#                          './hllhc_14/line_and_particle.json')
#rtol_100turns = 1e-9; atol_100turns=8e-11

# fname_line_particles = test_data_folder.joinpath(
#                  './sps_w_spacecharge/line_no_spacecharge_and_particle.json')
# rtol_100turns = 1e-9; atol_100turns=3e-11

# fname_line_particles = test_data_folder.joinpath(
#                    './sps_w_spacecharge/line_with_spacecharge_and_particle.json')
# rtol_100turns = 2e-8; atol_100turns=7e-9

short_test = False
num_turns = int(100)

####################
# Choose a context #
####################

#n_part = 8000
#context = xo.ContextCpu(omp_num_threads=8)

n_part = 200
context = xo.ContextCpu(omp_num_threads=0)

# n_part = 20000
# context = xo.ContextCupy()

#n_part = 20000
#context = xo.ContextPyopencl('0.0')

#############
# Load file #
#############

with open(fname_line_particles, 'r') as fid:
    input_data = json.load(fid)

##############
# Get a line #
##############

print('Import line')
line= xt.Line.from_dict(input_data['line'])

##################
# Build TrackJob #
##################

print('Build tracker')
line.build_tracker(_context=context)
line.reset_s_at_end_turn = False

######################
# Get some particles #
######################

print('Import particles')
line.particle_ref = xt.Particles(**input_data['particle'])

# Go from one particle to many particles

particles = line.build_particles(_context=context,
    x=np.arange(-1e-4, 1e-4, 2e-4/n_part),
    y=np.arange(-2e-4, 2e-4, 4e-4/n_part))

#########
# Track #
#########
particles_before_tracking = particles.copy(_context=xo.ContextCpu())
print('Track!')
print(f'context: {line._buffer.context}')
line.track(particles, num_turns=num_turns, time=True)
print(f'Time {(line.time_last_track)*1000:.2f} ms')
print(f'Time {(line.time_last_track)*1e6/num_turns/n_part:.2f} us/part/turn')

###########################
# Check against ducktrack #
###########################

testline = dtk.TestLine.from_dict(input_data['line'])

ip_check = n_part//4*3

print(f'\nTest against ducktrack over {num_turns} turns on particle {ip_check}:')
vars_to_check = ['x', 'px', 'y', 'py', 'zeta', 'delta', 's']
part_dict = particles_before_tracking.to_dict()
part_to_check = {}
for kk, vv in part_dict.items():
    if hasattr(vv, '__iter__') and not kk.startswith('_'):
        part_to_check[kk] = part_dict[kk][ip_check]
    else:
        part_to_check[kk] = part_dict[kk]

dtk_part = dtk.TestParticles(**part_to_check)


for iturn in range(num_turns):
    print(f'turn {iturn}/{num_turns}', end='\r', flush=True)
    testline.track(dtk_part)

for vv in vars_to_check:
    dtk_value = getattr(dtk_part, vv)
    xt_value = context.nparray_from_context_array(
                        getattr(particles, vv)[ip_check])
    passed = np.isclose(xt_value, dtk_value,
                        rtol=rtol_100turns, atol=atol_100turns)
    if not passed:
        print(f'Not passed on var {vv}!\n'
              f'    dtk:    {dtk_value: .7e}\n'
              f'    xtrack: {xt_value: .7e}\n')
        raise ValueError
    else:
        print(f'Passed on var {vv}!\n'
              f'    dtk:    {dtk_value: .7e}\n'
              f'    xtrack: {xt_value: .7e}\n')



import json

import numpy as np
import xtrack as xt
import xobjects as xo

with open('../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json',
            'r') as fid:
    dct = json.load(fid)
line = xt.Line.from_dict(dct)
line.cycle('ip1', inplace=True)

line.build_tracker()

tw_before = line.twiss()

GreaterThan = xt.GreaterThan
LessThan = xt.LessThan

opt = line.match(
    solve=False,
    start='ip1', end='ip1.l1',
    init=tw_before, init_at=xt.START,
    vary=[
        xt.VaryList(['kqtf.b1', 'kqtd.b1'], step=1e-8),
    ],
    targets=[
        xt.TargetRelPhaseAdvance('mux', GreaterThan(62.26),
                                 start='ip1', end='ip1.l1', tol=1e-4),
        xt.TargetRelPhaseAdvance('mux', LessThan(62.27),
                                 start='ip1', end='ip1.l1', tol=1e-4),
        xt.TargetRelPhaseAdvance('muy', LessThan(60.28),
                                  start='ip1', end='ip1.l1', tol=1e-4),
    ]
)

opt.target_status()

opt.targets[0].freeze()
opt.targets[1].freeze()
opt.target_status()
opt.solve()

tt = line.twiss()
assert np.isclose(tt['mux'][-1], 62.31, atol=1e-4)
assert tt['muy'][-1] < 60.281

opt.targets[0].unfreeze()
opt.targets[1].unfreeze()
opt.target_status()
opt.solve()
tt = line.twiss()
assert tt.mux[-1] > 62.26 - 1e-3
assert tt.mux[-1] < 62.27 + 1e-3
assert tt.muy[-1] < 60.281


#!end-doc-part



import xtrack as xt
import numpy as np

env1 = xt.Environment()
env1.vars.default_to_zero  = True
line1 = env1.new_line(components=[
    env1.new('qq1_thick', xt.Quadrupole, length=1., k1='kk', at=10),
    env1.new('qq1_thin', xt.Quadrupole, length=1., k1='kk', at=20),
    env1.new('qq_shared_thick', xt.Quadrupole, length=1., k1='kk', at=30),
    env1.new('qq_shared_thin', xt.Quadrupole, length=1., k1='kk', at=40),
])
line1.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(slicing=xt.Teapot(2, mode='thick'), name='qq1_thick'),
        xt.Strategy(slicing=xt.Teapot(2, mode='thin'), name='qq1_thin'),
        xt.Strategy(slicing=xt.Teapot(2, mode='thick'), name='qq_shared_thick'),
        xt.Strategy(slicing=xt.Teapot(2, mode='thin'), name='qq_shared_thin'),
    ])

env2 = xt.Environment()
env2.vars.default_to_zero  = True
line2 = env2.new_line(components=[
    env2.new('qq2_thick', xt.Quadrupole, length=1., k1='kk', at=10),
    env2.new('qq2_thin', xt.Quadrupole, length=1., k1='kk', at=20),
    env2.new('qq_shared_thick', xt.Quadrupole, length=1., k1='kk', at=30),
    env2.new('qq_shared_thin', xt.Quadrupole, length=1., k1='kk', at=40),
])
line2.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(slicing=xt.Teapot(2, mode='thick'), name='qq2_thick'),
        xt.Strategy(slicing=xt.Teapot(2, mode='thin'), name='qq2_thin'),
        xt.Strategy(slicing=xt.Teapot(2, mode='thick'), name='qq_shared_thick'),
        xt.Strategy(slicing=xt.Teapot(2, mode='thin'), name='qq_shared_thin'),
    ])

# Merge the line in one environment
env = xt.Environment(lines={'line1': line1, 'line2': line2})

tt1 = env.line1.get_table()
tt2 = env.line2.get_table()

tt1.show(cols='name element_type parent_name')
# name                          element_type         parent_name
# drift_1                       Drift                None
# qq1_thick_entry               Marker               None
# qq1_thick..0                  ThickSliceQuadrupole qq1_thick
# qq1_thick..1                  ThickSliceQuadrupole qq1_thick
# qq1_thick_exit                Marker               None
# drift_2                       Drift                None
# qq1_thin_entry                Marker               None
# drift_qq1_thin..0             DriftSliceQuadrupole qq1_thin
# qq1_thin..0                   ThinSliceQuadrupole  qq1_thin
# drift_qq1_thin..1             DriftSliceQuadrupole qq1_thin
# qq1_thin..1                   ThinSliceQuadrupole  qq1_thin
# drift_qq1_thin..2             DriftSliceQuadrupole qq1_thin
# qq1_thin_exit                 Marker               None
# drift_3                       Drift                None
# qq_shared_thick_entry         Marker               None
# qq_shared_thick..0/line1      ThickSliceQuadrupole qq_shared_thick/line1
# qq_shared_thick..1/line1      ThickSliceQuadrupole qq_shared_thick/line1
# qq_shared_thick_exit          Marker               None
# drift_4                       Drift                None
# qq_shared_thin_entry          Marker               None
# drift_qq_shared_thin..0/line1 DriftSliceQuadrupole qq_shared_thin/line1
# qq_shared_thin..0/line1       ThinSliceQuadrupole  qq_shared_thin/line1
# drift_qq_shared_thin..1/line1 DriftSliceQuadrupole qq_shared_thin/line1
# qq_shared_thin..1/line1       ThinSliceQuadrupole  qq_shared_thin/line1
# drift_qq_shared_thin..2/line1 DriftSliceQuadrupole qq_shared_thin/line1
# qq_shared_thin_exit           Marker               None

tt2.show(cols='name element_type parent_name')
# name                          element_type         parent_name
# drift_5                       Drift                None
# qq2_thick_entry               Marker               None
# qq2_thick..0                  ThickSliceQuadrupole qq2_thick
# qq2_thick..1                  ThickSliceQuadrupole qq2_thick
# qq2_thick_exit                Marker               None
# drift_6                       Drift                None
# qq2_thin_entry                Marker               None
# drift_qq2_thin..0             DriftSliceQuadrupole qq2_thin
# qq2_thin..0                   ThinSliceQuadrupole  qq2_thin
# drift_qq2_thin..1             DriftSliceQuadrupole qq2_thin
# qq2_thin..1                   ThinSliceQuadrupole  qq2_thin
# drift_qq2_thin..2             DriftSliceQuadrupole qq2_thin
# qq2_thin_exit                 Marker               None
# drift_7                       Drift                None
# qq_shared_thick_entry         Marker               None
# qq_shared_thick..0/line2      ThickSliceQuadrupole qq_shared_thick/line2
# qq_shared_thick..1/line2      ThickSliceQuadrupole qq_shared_thick/line2
# qq_shared_thick_exit          Marker               None
# drift_8                       Drift                None
# qq_shared_thin_entry          Marker               None
# drift_qq_shared_thin..0/line2 DriftSliceQuadrupole qq_shared_thin/line2
# qq_shared_thin..0/line2       ThinSliceQuadrupole  qq_shared_thin/line2
# drift_qq_shared_thin..1/line2 DriftSliceQuadrupole qq_shared_thin/line2
# qq_shared_thin..1/line2       ThinSliceQuadrupole  qq_shared_thin/line2
# drift_qq_shared_thin..2/line2 DriftSliceQuadrupole qq_shared_thin/line2
# qq_shared_thin_exit           Marker               None
# _end_point                                         None

assert np.all(tt1.name == np.array([
       'drift_1', 'qq1_thick_entry', 'qq1_thick..0', 'qq1_thick..1',
       'qq1_thick_exit', 'drift_2', 'qq1_thin_entry', 'drift_qq1_thin..0',
       'qq1_thin..0', 'drift_qq1_thin..1', 'qq1_thin..1',
       'drift_qq1_thin..2', 'qq1_thin_exit', 'drift_3',
       'qq_shared_thick_entry', 'qq_shared_thick..0/line1',
       'qq_shared_thick..1/line1', 'qq_shared_thick_exit', 'drift_4',
       'qq_shared_thin_entry', 'drift_qq_shared_thin..0/line1',
       'qq_shared_thin..0/line1', 'drift_qq_shared_thin..1/line1',
       'qq_shared_thin..1/line1', 'drift_qq_shared_thin..2/line1',
       'qq_shared_thin_exit', '_end_point']))

assert np.all(tt1.element_type == np.array([
       'Drift', 'Marker', 'ThickSliceQuadrupole', 'ThickSliceQuadrupole',
       'Marker', 'Drift', 'Marker', 'DriftSliceQuadrupole',
       'ThinSliceQuadrupole', 'DriftSliceQuadrupole',
       'ThinSliceQuadrupole', 'DriftSliceQuadrupole', 'Marker', 'Drift',
       'Marker', 'ThickSliceQuadrupole', 'ThickSliceQuadrupole', 'Marker',
       'Drift', 'Marker', 'DriftSliceQuadrupole', 'ThinSliceQuadrupole',
       'DriftSliceQuadrupole', 'ThinSliceQuadrupole',
       'DriftSliceQuadrupole', 'Marker', '']))

assert np.all(tt1.parent_name == np.array([
       None, None, 'qq1_thick', 'qq1_thick', None, None, None, 'qq1_thin',
       'qq1_thin', 'qq1_thin', 'qq1_thin', 'qq1_thin', None, None, None,
       'qq_shared_thick/line1', 'qq_shared_thick/line1', None, None, None,
       'qq_shared_thin/line1', 'qq_shared_thin/line1',
       'qq_shared_thin/line1', 'qq_shared_thin/line1',
       'qq_shared_thin/line1', None, None]))

assert np.all(tt2.name == np.array([
       'drift_5', 'qq2_thick_entry', 'qq2_thick..0', 'qq2_thick..1',
       'qq2_thick_exit', 'drift_6', 'qq2_thin_entry', 'drift_qq2_thin..0',
       'qq2_thin..0', 'drift_qq2_thin..1', 'qq2_thin..1',
       'drift_qq2_thin..2', 'qq2_thin_exit', 'drift_7',
       'qq_shared_thick_entry', 'qq_shared_thick..0/line2',
       'qq_shared_thick..1/line2', 'qq_shared_thick_exit', 'drift_8',
       'qq_shared_thin_entry', 'drift_qq_shared_thin..0/line2',
       'qq_shared_thin..0/line2', 'drift_qq_shared_thin..1/line2',
       'qq_shared_thin..1/line2', 'drift_qq_shared_thin..2/line2',
       'qq_shared_thin_exit', '_end_point']))

assert np.all(tt2.element_type == np.array([
       'Drift', 'Marker', 'ThickSliceQuadrupole', 'ThickSliceQuadrupole',
       'Marker', 'Drift', 'Marker', 'DriftSliceQuadrupole',
       'ThinSliceQuadrupole', 'DriftSliceQuadrupole',
       'ThinSliceQuadrupole', 'DriftSliceQuadrupole', 'Marker', 'Drift',
       'Marker', 'ThickSliceQuadrupole', 'ThickSliceQuadrupole', 'Marker',
       'Drift', 'Marker', 'DriftSliceQuadrupole', 'ThinSliceQuadrupole',
       'DriftSliceQuadrupole', 'ThinSliceQuadrupole',
       'DriftSliceQuadrupole', 'Marker', '']))

assert np.all(tt2.parent_name == np.array([
       None, None, 'qq2_thick', 'qq2_thick', None, None, None, 'qq2_thin',
       'qq2_thin', 'qq2_thin', 'qq2_thin', 'qq2_thin', None, None, None,
       'qq_shared_thick/line2', 'qq_shared_thick/line2', None, None, None,
       'qq_shared_thin/line2', 'qq_shared_thin/line2',
       'qq_shared_thin/line2', 'qq_shared_thin/line2',
       'qq_shared_thin/line2', None, None]))

assert 'qq1_thick' in env.element_dict
assert 'qq1_thin' in env.element_dict
assert 'qq_shared_thick/line1' in env.element_dict
assert 'qq_shared_thin/line1' in env.element_dict
assert 'qq2_thick' in env.element_dict
assert 'qq2_thin' in env.element_dict
assert 'qq_shared_thick/line2' in env.element_dict
assert 'qq_shared_thin/line2' in env.element_dict

line1['kk'] = 1e-1
line2['kk'] = 1e-1
env['kk'] = 1e-1

particle_ref = xt.Particles(p0c=7e12)
line1.particle_ref = particle_ref
line2.particle_ref = particle_ref
env.line1.particle_ref = particle_ref
env.line2.particle_ref = particle_ref

tw1 = line1.twiss(betx=1, bety=2)
tw2 = line2.twiss(betx=1, bety=2)
tw1i = env.line1.twiss(betx=1, bety=2)
tw2i = env.line2.twiss(betx=1, bety=2)

assert np.allclose(tw1.s, tw1i.s, atol=0, rtol=1e-15)
assert np.allclose(tw1.betx, tw1i.betx, atol=0, rtol=1e-15)
assert np.allclose(tw1.bety, tw1i.bety, atol=0, rtol=1e-15)

assert np.allclose(tw2.s, tw2i.s, atol=0, rtol=1e-15)
assert np.allclose(tw2.betx, tw2i.betx, atol=0, rtol=1e-15)
assert np.allclose(tw2.bety, tw2i.bety, atol=0, rtol=1e-15)



import xtrack as xt
import numpy as np

from scipy.constants import hbar
from scipy.constants import electron_volt
from scipy.constants import c as clight

env = xt.load_madx_lattice('../../test_data/sps_thick/sps.seq')
env.vars.load_madx('../../test_data/sps_thick/lhc_q20.str')
line = env.sps

line['actcse.31632'].voltage = 4.2e+08
line['actcse.31632'].frequency = 3e6
line['actcse.31632'].lag = 180.

tt = line.get_table()


line.particle_ref = xt.Particles(energy0=20e9, mass0=xt.ELECTRON_MASS_EV)
env.particle_ref = line.particle_ref

import wiggler as wgl

# Wiggler parameters
k0_wig = 5e-3
tilt_rad = np.pi/2

lenwig = 25
numperiods = 20
lambdawig = lenwig / numperiods

wig = wgl.Wiggler(period=lambdawig, amplitude=k0_wig, num_periods=numperiods,
                  angle_rad=tilt_rad, scheme='121a')

tt = line.get_table()
wig_elems = []
for name, element in wig.wiggler_dict.items():
    env.elements[name] = element['element']
    wig_elems.append(name)

wig_line = env.new_line(components=[
                        env.new('s.wig', 'Marker'),
                        wig_elems,
                        env.new('e.wig', 'Marker'),
])


line.insert(wig_line, anchor='start', at=1, from_='qd.31710@end')

tt = line.get_table()
tw4d_thick = line.twiss4d()
tw6d_thick = line.twiss()

env['sps_thick'] = env.sps.copy(shallow=True)

line.discard_tracker()
slicing_strategies = [
    xt.Strategy(slicing=xt.Teapot(1)),  # Default
    xt.Strategy(slicing=xt.Teapot(2), element_type=xt.Bend),
    xt.Strategy(slicing=xt.Teapot(2), element_type=xt.RBend),
    xt.Strategy(slicing=xt.Teapot(8), element_type=xt.Quadrupole),
    xt.Strategy(slicing=xt.Teapot(20), name='mwp.*'),
]
line.slice_thick_elements(slicing_strategies)

tw4d = line.twiss4d(radiation_integrals=True)
tw6d = line.twiss()

line.configure_radiation(model='mean')

tw_rad = line.twiss(eneloss_and_damping=True, strengths=True)

print('ex rad int:', tw4d.rad_int_eq_gemitt_x)
print('ex Chao:   ', tw_rad.eq_gemitt_x)
print('ey rad int:', tw4d.rad_int_eq_gemitt_y)
print('ey Chao:   ', tw_rad.eq_gemitt_y)

print('damping rate x [s^-1] rad int:   ', tw4d.rad_int_damping_constant_x_s)
print('damping rate x [s^-1] eigenval:  ', tw_rad.damping_constants_s[0])
print('damping rate y [s^-1] rad int:   ', tw4d.rad_int_damping_constant_y_s)
print('damping rate y [s^-1] eigenval:  ', tw_rad.damping_constants_s[1])
print('damping rate z [s^-1] rad int:   ', tw4d.rad_int_damping_constant_zeta_s)
print('damping rate z [s^-1] eigenval:  ', tw_rad.damping_constants_s[2])


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time                          #!skip-doc
import numpy as np

import xtrack as xt

######################################
# Create a dummy collimation process #
######################################

class DummyInteractionProcess:
    '''
    I kill some particles. I kick some others by an given angle
    and I generate some secondaries with the opposite angles.
    '''
    def __init__(self, fraction_lost, fraction_secondary, length, kick_x):

        self.fraction_lost = fraction_lost
        self.fraction_secondary = fraction_secondary
        self.kick_x = kick_x
        self.length = length

        self.drift = xt.Drift(length=self.length)


    def interact(self, particles):

        self.drift.track(particles)

        n_part = particles._num_active_particles

        # Kill some particles
        mask_kill = np.random.uniform(size=n_part) < self.fraction_lost
        particles.state[:n_part][mask_kill] = -1 # special flag`


        # Generate some more particles
        mask_secondary = np.random.uniform(size=n_part) < self.fraction_secondary
        n_products = np.sum(mask_secondary)
        if n_products>0:
            products = {
                's': particles.s[:n_part][mask_secondary],
                'x': particles.x[:n_part][mask_secondary] + self.length * (
                    particles.px[:n_part][mask_secondary] + self.kick_x),
                'px': particles.px[:n_part][mask_secondary] + self.kick_x,
                'y': particles.y[:n_part][mask_secondary],
                'py': particles.py[:n_part][mask_secondary],
                'zeta': particles.zeta[:n_part][mask_secondary],
                'delta': particles.delta[:n_part][mask_secondary],

                'mass_ratio': particles.x[:n_part][mask_secondary] *0 + 1.,
                'charge_ratio': particles.x[:n_part][mask_secondary] *0 + 1.,

                'parent_particle_id': particles.particle_id[:n_part][mask_secondary],
                'at_element': particles.at_element[:n_part][mask_secondary],
                'at_turn': particles.at_turn[:n_part][mask_secondary],
                }
        else:
            products = None

        return products

############################################################
# Create a beam interaction from the process defined above #
############################################################

# kick_x is large so that the particles are lost on the first aperture element
interaction_process=DummyInteractionProcess(length=1., kick_x=4e-2,
                                            fraction_lost=0.0,
                                            fraction_secondary=0.2)
collimator = xt.BeamInteraction(length=interaction_process.length,
                                      interaction_process=interaction_process)

########################################################
# Create a line including the collimator defined above #
########################################################

line = xt.Line(elements=[
    xt.Drift(length=2.),
    xt.LimitEllipse(a=2e-2, b=2e-2),
    collimator,
    xt.LimitEllipse(a=2e-2, b=2e-2),
    xt.Drift(length=10.),
    ])

#################
# Build tracker #
#################

line.build_tracker()
line.config.XTRACK_GLOBAL_XY_LIMIT = 1e3

##########################
# Build particles object #
##########################

# We prepare empty slots to store the product particles that will be
# generated during the tracking.
particles = xt.Particles(
        _capacity=200000,
        x=np.zeros(100000))

#########
# Track #
#########

t1 = time.time()                                          #!skip-doc
line.track(particles)
t2 = time.time()                                          #!skip-doc

                                                          #!skip-doc
print(f'{t2-t1=:.2f}')                                    #!skip-doc

############################
# Loss location refinement #
############################

part_before = particles.copy()

loss_loc_refinement = xt.LossLocationRefinement(line,
                                            n_theta = 360,
                                            r_max = 0.5, # m
                                            dr = 50e-6,
                                            ds = 0.05,
                                            save_refine_lines=True)

loss_loc_refinement.refine_loss_location(particles)

assert np.all(part_before.s == particles.s)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import os

copyright_statement = (
"""copyright ###############################
This file is part of the Xtrack Package. 
Copyright (c) CERN, 2021.                
#########################################""")

config = [
    {'extension': '.py', 'comment_char': '#'},
    {'extension': '.h', 'comment_char': '//'},
]

for cc in config:
    extension =  cc['extension']
    comment_char = cc['comment_char']

    cpright_lines = [comment_char + ' ' + line + " " + comment_char + '\n'
                        for line in copyright_statement.splitlines()
                        ] + ['\n']

    for root, dirs, files in os.walk("./"):
        for fname in files:
            if fname.endswith(extension):
                file = os.path.join(root, fname)
                print(file)
                with open(file, 'r') as fid:
                    lines = fid.readlines()
                # Remove copyright statement if present
                if (len(lines) > 0 and
                    lines[0].startswith(comment_char + ' ' + 'copyright ##')):
                    for ill, ll in enumerate(lines):
                        assert ll.startswith(comment_char)
                        if ll.startswith(comment_char + ' ' + '########'):
                            end_cpright = ill
                            break
                    assert lines[end_cpright+1] == '\n'
                    lines = lines[end_cpright+2:]

                lines = cpright_lines + lines

                with open(file, 'w') as fid:
                    fid.writelines(lines)



import xtrack as xt

# We create an environment and define a quadrupole and a drift
env = xt.Environment()
env['kq'] = 0.1
env.new('mq', xt.Quadrupole, length=0.3, k1='kq')
env.new('dd', xt.Drift, length=1)

#####################
# Repeated elements #
#####################

# We create a line with repeated elements
line = env.new_line(components=['mq', 'dd', 'mq', 'dd'])

line.get_table(attr=True).cols['name s rot_s_rad']
# is:
# name                   s     rot_s_rad
# mq::0                  0             0
# dd::0                0.3             0
# mq::1                1.3             0
# dd::1                1.6             0
# _end_point           2.6             0

# Here 'mq::0' and 'mq::1' are actually the same element. Any modification on
# 'mq' is seen directly on 'mq::0' and 'mq::1'. For example, we  tilt mq by 1 mrad:
line['mq'].rot_s_rad = 1e-3

line.get_table(attr=True).cols['name s rot_s_rad']
# is:
# name                   s     rot_s_rad
# mq::0                  0         0.001
# dd::0                0.3             0
# mq::1                1.3         0.001
# dd::1                1.6             0
# _end_point           2.6             0

############
# Replicas #
############

# Replicas behave in the same way as repeated elements, but allow assigning
# a different name to each replica. For example:
env.new('my_mq_1', 'mq', mode='replica')
env.new('my_mq_2', 'mq', mode='replica')

line = env.new_line(components=['my_mq_1', 'dd', 'my_mq_2', 'dd'])

# Here 'mq_1' and 'mq::1' are actually the same element. Any modification on
# 'mq' is seen directly on 'mq::0' and 'mq::1'. For example, we  set the tilt of
# mq by 3 mrad:
line['mq'].rot_s_rad = 3e-3

line.get_table(attr=True).cols['name s rot_s_rad']
# is:
# Table: 5 rows, 3 cols
# name                   s     rot_s_rad
# my_mq_1                0         0.001
# dd::0                0.3             0
# my_mq_2              1.3         0.001
# dd::1                1.6             0
# _end_point           2.6             0

##################
# Element clones #
##################

# Element clones are different from replicas and repeated elements. They are
# actual copies of the element, which inherit the deferred expressions controlling
# their attributes from the original element. For example:

env.new('mq_clone_1', 'mq', mode='clone')
env.new('mq_clone_2', 'mq', mode='clone')

line = env.new_line(components=['mq_clone_1', 'dd', 'mq_clone_2', 'dd'])

line['mq'].get_expr('k1') # is 'kq'
line['mq_clone_1'].get_expr('k1') # is 'kq'
line['mq_clone_2'].get_expr('k1') # is 'kq'

# When changing the value of 'kq', all three elements are affected
env['kq'] = 0.2
line['mq'].k1 # is 0.2
line['mq_clone_1'].k1 # is 0.2
line['mq_clone_2'].k1 # is 0.2

# Note instead that if we alter the expression controlling the attribute of mq after
# its creation, the clones are not affected. For example:
line['mq'].k1 = '2 * kq'

line['mq'].get_expr('k1')         # is '2 * kq'
line['mq_clone_1'].get_expr('k1') # is 'kq'
line['mq_clone_2'].k1             # is 'kq'

# Clones allow for example, specifying different values for tilts and offsets
# for different elements. For example:

line['mq_clone_1'].rot_s_rad = 2e-3
line['mq_clone_2'].rot_s_rad = 3e-3

line.get_table(attr=True).cols['name s k1l rot_s_rad']
# is:
# Table: 5 rows, 4 cols
# name                   s           k1l     rot_s_rad
# mq_clone_1             0          0.03         0.002
# dd::0                0.3             0             0
# mq_clone_2           1.3          0.03         0.003
# dd::1                1.6             0             0
# _end_point           2.6             0             0

##########################################
# Replace repleted elements and replicas #
##########################################

# The line provides methods to automatically replace repeated elements and replicas
# with clones. For example:

line = env.new_line(components=['mq', 'dd', 'mq', 'dd', 'my_mq_1'])
line.get_table(attr=True).cols['name s isreplica']
# is:
# name                   s isreplica
# mq::0                  0     False
# dd::0                0.3     False
# mq::1                1.3     False
# dd::1                1.6     False
# my_mq_1              2.6      True
# _end_point           2.9     False

line.replace_all_repeated_elements()
line.replace_all_replicas()

line.get_table(attr=True).cols['name s isreplica']
# is:
# name                   s isreplica
# mq.0                   0     False
# dd.0                 0.3     False
# mq.1                 1.3     False
# dd.1                 1.6     False
# my_mq_1              2.6     False
# _end_point           2.9     False



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib

import xobjects as xo
from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts(excluding='ContextPyopencl')
def test_instability_cpu_gpu(test_context):

    import time
    import numpy as np
    from scipy import constants

    protonMass = constants.value("proton mass energy equivalent in MeV") * 1e6
    from scipy.stats import linregress
    from scipy.signal import hilbert

    import xtrack as xt
    import xpart as xp
    xp.enable_pyheadtail_interface()

    from PyHEADTAIL.particles.generators import generate_Gaussian6DTwiss
    from PyHEADTAIL.particles.slicing import UniformBinSlicer
    from PyHEADTAIL.impedances.wakes import WakeTable, WakeField
    from PyHEADTAIL.feedback.transverse_damper import TransverseDamper
    from PyHEADTAIL.trackers.transverse_tracking import TransverseSegmentMap
    from PyHEADTAIL.trackers.longitudinal_tracking import LinearMap
    from PyHEADTAIL.trackers.detuners import ChromaticitySegment, AmplitudeDetuningSegment

    nTurn = 5000  # int(1E4)
    bunch_intensity = 1.8e11
    n_macroparticles = int(1e4)
    energy = 7e3  # [GeV]
    gamma = energy * 1e9 / protonMass
    betar = np.sqrt(1 - 1 / gamma ** 2)
    normemit = 1.8e-6
    beta_x = 68.9
    beta_y = 70.34
    Q_x = 0.31
    Q_y = 0.32
    chroma = -5.0  # 10.0
    sigma_4t = 1.2e-9
    sigma_z = sigma_4t / 4.0 * constants.c
    momentumCompaction = 3.483575072011584e-04
    eta = momentumCompaction - 1.0 / gamma ** 2
    voltage = 12.0e6
    h = 35640
    p0 = constants.m_p * betar * gamma * constants.c
    Q_s = np.sqrt(constants.e * voltage * eta * h / (2 * np.pi * betar * constants.c * p0))
    circumference = 26658.883199999
    averageRadius = circumference / (2 * np.pi)
    sigma_delta = Q_s * sigma_z / (averageRadius * eta)
    beta_s = sigma_z / sigma_delta
    emit_s = 4 * np.pi * sigma_z * sigma_delta * p0 / constants.e  # eVs for PyHEADTAIL

    n_slices_wakes = 200
    limit_z = 3 * sigma_z
    wake_folder = pathlib.Path(
        __file__).parent.joinpath('../examples/pyheadtail_interface/wakes').absolute()
    wakefile = wake_folder.joinpath(
        "wakeforhdtl_PyZbase_Allthemachine_7000GeV_B1_2021_TeleIndex1_wake.dat")
    slicer_for_wakefields = UniformBinSlicer(n_slices_wakes, z_cuts=(-limit_z, limit_z))
    waketable = WakeTable(
        wakefile, ["time", "dipole_x", "dipole_y", "quadrupole_x", "quadrupole_y"]
    )
    wake_field = WakeField(slicer_for_wakefields, waketable)
    wake_field.needs_cpu = True
    wake_field.needs_hidden_lost_particles = True

    damping_time = 7000  # 33.
    damper = TransverseDamper(dampingrate_x=damping_time, dampingrate_y=damping_time)
    damper.needs_cpu = True
    damper.needs_hidden_lost_particles = True
    i_oct = 15.
    det_xx = 1.4e5 * i_oct / 550.0  # from PTC with ATS optics, telescopic factor 1.0
    det_xy = -1.0e5 * i_oct / 550.0

    # expected octupole threshold with damper is 273A according to https://indico.cern.ch/event/902528/contributions/3798807/attachments/2010534/3359300/20200327_RunIII_stability_NMounet.pdf
    # expected growth rate with damper but without octupole is ~0.3 [$10^{-4}$/turn] (also according to Nicolas' presentation)

    checkTurn = 1000

    ########### PyHEADTAIL part ##############

    particles = generate_Gaussian6DTwiss(
        macroparticlenumber=n_macroparticles,
        intensity=bunch_intensity,
        charge=constants.e,
        mass=constants.m_p,
        circumference=circumference,
        gamma=gamma,
        alpha_x=0.0,
        alpha_y=0.0,
        beta_x=beta_x,
        beta_y=beta_y,
        beta_z=beta_s,
        epsn_x=normemit,
        epsn_y=normemit,
        epsn_z=emit_s,
    )

    x0 = np.copy(particles.x)
    px0 = np.copy(particles.xp)
    y0 = np.copy(particles.y)
    py0 = np.copy(particles.yp)
    zeta0 = np.copy(particles.z)
    delta0 = np.copy(particles.dp)

    chromatic_detuner = ChromaticitySegment(dQp_x=chroma, dQp_y=0.0)
    transverse_detuner = AmplitudeDetuningSegment(
        dapp_x=det_xx * p0,
        dapp_y=det_xx * p0,
        dapp_xy=det_xy * p0,
        dapp_yx=det_xy * p0,
        alpha_x=0.0,
        beta_x=beta_x,
        alpha_y=0.0,
        beta_y=beta_y,
    )
    arc_transverse = TransverseSegmentMap(
        alpha_x_s0=0.0,
        beta_x_s0=beta_x,
        D_x_s0=0.0,
        alpha_x_s1=0.0,
        beta_x_s1=beta_x,
        D_x_s1=0.0,
        alpha_y_s0=0.0,
        beta_y_s0=beta_y,
        D_y_s0=0.0,
        alpha_y_s1=0.0,
        beta_y_s1=beta_y,
        D_y_s1=0.0,
        dQ_x=Q_x,
        dQ_y=Q_y,
        segment_detuners=[chromatic_detuner, transverse_detuner],
    )
    arc_longitudinal = LinearMap(
        alpha_array=[momentumCompaction], circumference=circumference, Q_s=Q_s
    )

    turns = np.arange(nTurn)
    x = np.zeros(nTurn, dtype=float)
    t_pyht_start = time.time()
    for turn in range(nTurn):
        time0 = time.time()
        arc_transverse.track(particles)
        arc_longitudinal.track(particles)
        time1 = time.time()
        wake_field.track(particles)
        time2 = time.time()
        damper.track(particles)
        time3 = time.time()
        x[turn] = np.mean(particles.x)
        if turn % 1000 == 0:
            print(
                f"PyHt - turn {turn}" )
    t_pyht_end = time.time()
    print(f'PyHT full loop: {t_pyht_end - t_pyht_start}s')

    x /= np.sqrt(normemit * beta_x / gamma / betar)
    iMin = 1000
    iMax = nTurn - 1000
    if iMin >= iMax:
        iMin = 0
        iMax = nTurn
    ampl = np.abs(hilbert(x))
    b, a, r, p, stderr = linregress(turns[iMin:iMax], np.log(ampl[iMin:iMax]))
    gr_pyht = b

    p_pht = particles

    ############ xsuite-PyHEADTAIL part (the WakeField instance is shared) ########################

    particles = xp.Particles(
        _context=test_context,
        circumference=circumference,
        particlenumber_per_mp=bunch_intensity / n_macroparticles,
        q0=1,
        mass0=protonMass,
        gamma0=gamma,
        x=np.zeros(2*len(x0))
    )

    np2ctx = test_context.nparray_to_context_array
    particles.x[::2] = np2ctx(x0)
    particles.px[::2] = np2ctx(px0)
    particles.y[::2] = np2ctx(y0)
    particles.py[::2] = np2ctx(py0)
    particles.zeta[::2] = np2ctx(zeta0)
    particles.delta[::2] = np2ctx(delta0)
    particles.state[1::2] = 0

    arc = xt.LineSegmentMap(
        _context=test_context,
        betx=beta_x,
        bety=beta_y,
        qx=Q_x,
        qy=Q_y,
        bets=beta_s,
        qs=Q_s,
        dqx=chroma,
        det_xx=det_xx,
        det_xy=det_xy,
        det_yy=det_xx,
        det_yx=det_xy,
        energy_ref_increment=0.0,
        energy_increment=0,
    )

    line = xt.Line(elements=[arc, wake_field, damper],
                       element_names=['arc', 'wake_field', 'damper'])
    line.build_tracker(_context=test_context)

    t_xt_start = time.time()
    turns = np.arange(nTurn)
    x = np.zeros(nTurn, dtype=float)
    for turn in range(nTurn):

        line.track(particles)

        x[turn] = np.average(particles.x[particles.state>0])
        if turn % 1000 == 0:
            print(
                f"PyHtXt - turn {turn}"
            )
    t_xt_end = time.time()

    print(f'Xt full loop: {t_xt_end - t_xt_start}s')

    x /= np.sqrt(normemit * beta_x / gamma / betar)
    iMin = 1000
    iMax = nTurn - 1000
    if iMin >= iMax:
        iMin = 0
        iMax = nTurn
    ampl = np.abs(hilbert(x))
    b, a, r, p, stderr = linregress(turns[iMin:iMax], np.log(ampl[iMin:iMax]))
    gr_xtpyht = b
    print(f"Growth rate {b*1E4} [$10^{-4}$/turn]")

    print(f'{gr_pyht=}, {gr_xtpyht=} {gr_pyht-gr_xtpyht=}')
    xo.assert_allclose(gr_xtpyht, gr_pyht, rtol=1e-3, atol=1e-100)

    xp.disable_pyheadtail_interface() # would stay enabled for following tests
                                      # called by pytest



import pathlib

from cpymad.madx import Madx

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()




@for_all_test_contexts
def test_chi(test_context):
    ''' Tests if particles are deflected in quadrupoles according to their chi value.'''

    quad1 = xt.Quadrupole(knl=[0,-0.005])
    ql    = xt.Line(
                elements      = [xt.Drift(length=10), quad1,    xt.Drift(length=100)],
                element_names = ['drift_0', 'quad_0', 'drift_1'])
    

    ql.build_tracker(_context = test_context)

    print(f"Test {test_context.__class__}")

    z0              = 82
    z               = 82
    m               = 192755492342.6663   # pb207
    m0              = 193687690162.6481
    ql.particle_ref = xp.Particles(p0c=6.8e12*z0 , q0=z0, mass0=m0)

    # reference species with effective rigidity
    part1           = ql.particle_ref.copy(_context=test_context)
    delta           = 0
    chi             = (z0/z)*(m/m0)
    part1.delta     = (1 + delta)/(chi) - 1

    # pb207
    part2              = ql.particle_ref.copy(_context=test_context)
    part2.chi          = (z0/z)*(m/m0)
    part2.charge_ratio = (z/z0)

    # 
    x0        = 0.0001
    part1.x   = x0
    part2.x   = x0

    ql.track(part1)
    ql.track(part2)

    xo.assert_allclose(part1.x[0], part2.x[0], rtol=0, atol=1e-6)


@for_all_test_contexts
def test_ions(test_context):

    mad = Madx(stdout=False)
    mad.call(str(test_data_folder.joinpath(
        'sps_ions/SPS_2021_Pb_ions_thin_test.seq')))
    mad.use('sps')
    mad.twiss()

    twmad_4d = mad.table.twiss.dframe()
    summad_4d = mad.table.summ.dframe()

    V_RF = 1.7e6 # V (control room definition, energy gain per charge)

    # I switch on one cavity
    charge = mad.sequence.sps.beam.charge
    mad.sequence.sps.elements['actcse.31632'].volt = V_RF/1e6 * charge
    mad.sequence.sps.elements['actcse.31632'].lag = 0
    mad.sequence.sps.elements['actcse.31632'].freq = 200.

    twmad_6d = mad.table.twiss.dframe()
    summad_6d = mad.table.summ.dframe()

    mad.emit()
    qs_mad = mad.table.emitsumm.qs[0]

    print(f"Test {test_context.__class__}")
    # Make xsuite line and line
    line = xt.Line.from_madx_sequence(mad.sequence.sps, deferred_expressions=True)
    line.particle_ref = xp.Particles(mass0=mad.sequence.sps.beam.mass*1e9,
                                    q0=mad.sequence.sps.beam.charge,
                                    gamma0=mad.sequence.sps.beam.gamma)

    xo.assert_allclose(line['actcse.31632'].voltage, V_RF, atol=1e-10)

    line.build_tracker()

    tw = line.twiss()

    xo.assert_allclose(tw.qs, qs_mad, rtol=0, atol=1e-6)
    xo.assert_allclose(tw.qx, float(summad_4d.q1.iloc[0]), rtol=0, atol=1e-5)
    xo.assert_allclose(tw.qy, float(summad_4d.q2.iloc[0]), rtol=0, atol=1e-5)
    xo.assert_allclose(tw.dqx - float(summad_6d.dq1.iloc[0]), 0, rtol=0, atol=0.5)
    xo.assert_allclose(tw.dqy - float(summad_6d.dq2.iloc[0]), 0, rtol=0, atol=0.5)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib
import json
import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt
import xfields as xf

############
# Settings #
############

fname_line = ('../../test_data/sps_w_spacecharge/'
                  'line_no_spacecharge_and_particle.json')

# Realistic settings (feasible only on GPU)
bunch_intensity = 1e11/3 # Need short bunch to avoid bucket non-linearity
sigma_z = 22.5e-2/3
nemitt_x=2.5e-6
nemitt_y=2.5e-6
n_part=int(1e6)
num_turns=32
nz_grid = 100
z_range = (-3*sigma_z, 3*sigma_z)

num_spacecharge_interactions = 540

mode = 'frozen'
mode = 'quasi-frozen'
#mode = 'pic'

#context = xo.ContextCpu()
context = xo.ContextCupy()
#context = xo.ContextPyopencl('0.0')

print(context)

############
# Get line #
############

with open(fname_line, 'r') as fid:
     input_data = json.load(fid)
line= xt.Line.from_dict(input_data['line'])
line.particle_ref = xt.Particles.from_dict(input_data['particle'])
line.build_tracker(_context=context, compile=False) # Move all elements in the same context

#############################################
# Install spacecharge interactions (frozen) #
#############################################

lprofile = xf.LongitudinalProfileQGaussian(
        number_of_particles=bunch_intensity,
        sigma_z=sigma_z,
        z0=0.,
        q_parameter=1.)

xf.install_spacecharge_frozen(line=line,
                   longitudinal_profile=lprofile,
                   nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                   sigma_z=sigma_z,
                   num_spacecharge_interactions=num_spacecharge_interactions)

##########################
# Configure space-charge #
##########################

if mode == 'frozen':
    pass # Already configured in line
elif mode == 'quasi-frozen':
    xf.replace_spacecharge_with_quasi_frozen(
                                    line,
                                    update_mean_x_on_track=True,
                                    update_mean_y_on_track=True)
elif mode == 'pic':
    pic_collection, all_pics = xf.replace_spacecharge_with_PIC(
        line=line,
        n_sigmas_range_pic_x=8,
        n_sigmas_range_pic_y=8,
        nx_grid=256, ny_grid=256, nz_grid=nz_grid,
        n_lims_x=7, n_lims_y=3,
        z_range=z_range)
else:
    raise ValueError(f'Invalid mode: {mode}')

##################
# Build trackers #
##################

line.build_tracker(_context=context)
line_sc_off = line.filter_elements(exclude_types_starting_with='SpaceCh')

######################
# Generate particles #
######################

import footprint
r_max_sigma = 5
N_r_footprint = 10
N_theta_footprint = 8
theta_min = 0.05 * np.pi / 2
theta_max = np.pi / 2 - 0.05 * np.pi / 2

x_norm_fp, y_norm_fp, r_footprint, theta_footprint = xp.generate_2D_polar_grid(
        r_range=(0.3, r_max_sigma),
        nr=N_r_footprint+1,
        theta_range=(theta_min, theta_max), ntheta=N_theta_footprint)
N_footprint = len(x_norm_fp)

particles_fp = line.build_particles(
            weight=0, # pure probe particles
            zeta=0, delta=0,
            x_norm=x_norm_fp, px_norm=0,
            y_norm=y_norm_fp, py_norm=0,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y)

# I add explicitly a probe particle at1.5 sigma
particle_probe = line.build_particles(
            weight=0, # pure probe particles
            zeta=0, delta=0,
            x_norm=1.5, px_norm=0,
            y_norm=1.5, py_norm=0,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y)

particles_gaussian = xp.generate_matched_gaussian_bunch(
         num_particles=n_part, total_intensity_particles=bunch_intensity,
         nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
         line=line_sc_off)

particles = xt.Particles.merge(
                          [particles_fp, particle_probe, particles_gaussian])

particles_0 = particles.copy()

#########
# Track #
#########

ctx2arr = context.nparray_from_context_array
x_tbt = np.zeros((N_footprint, num_turns), dtype=np.float64)
y_tbt = np.zeros((N_footprint, num_turns), dtype=np.float64)
for ii in range(num_turns):
    print(f'Turn: {ii}', end='\r', flush=True)
    x_tbt[:, ii] = ctx2arr(particles.x[:N_footprint]).copy()
    y_tbt[:, ii] = ctx2arr(particles.y[:N_footprint]).copy()
    line.track(particles)

tw = line_sc_off.twiss(at_elements=[0])

######################
# Frequency analysis #
######################

import nafflib

xy_norm = np.zeros((N_r_footprint + 1, N_theta_footprint, 2), dtype=np.float64)
xy_norm[:, :, 0] = x_norm_fp.reshape((N_r_footprint + 1, N_theta_footprint))
xy_norm[:, :, 1] = y_norm_fp.reshape((N_r_footprint + 1, N_theta_footprint))

Qx = np.zeros(N_footprint)
Qy = np.zeros(N_footprint)

for i_part in range(N_footprint):
    Qx[i_part] = nafflib.get_tune(x_tbt[i_part, :])
    Qy[i_part] = nafflib.get_tune(y_tbt[i_part, :])

Qxy_fp = np.zeros_like(xy_norm)

Qxy_fp[:, :, 0] = np.reshape(Qx, Qxy_fp[:, :, 0].shape)
Qxy_fp[:, :, 1] = np.reshape(Qy, Qxy_fp[:, :, 1].shape)


###############################
# Tune shift from single turn #
###############################

p_probe_before = particles_0.filter(
        particles_0.particle_id == N_footprint).to_dict()

line.track(particles_0)

p_probe_after = particles_0.filter(
        particles_0.particle_id == N_footprint).to_dict()

betx = tw['betx'][0]
alfx = tw['alfx'][0]
phasex_0 = np.angle(p_probe_before['x'] / np.sqrt(betx) -
                   1j*(p_probe_before['x'] * alfx / np.sqrt(betx) +
                       p_probe_before['px'] * np.sqrt(betx)))[0]
phasex_1 = np.angle(p_probe_after['x'] / np.sqrt(betx) -
                   1j*(p_probe_after['x'] * alfx / np.sqrt(betx) +
                       p_probe_after['px'] * np.sqrt(betx)))[0]
bety = tw['bety'][0]
alfy = tw['alfy'][0]
phasey_0 = np.angle(p_probe_before['y'] / np.sqrt(bety) -
                   1j*(p_probe_before['y'] * alfy / np.sqrt(bety) +
                       p_probe_before['py'] * np.sqrt(bety)))[0]
phasey_1 = np.angle(p_probe_after['y'] / np.sqrt(bety) -
                   1j*(p_probe_after['y'] * alfy / np.sqrt(bety) +
                       p_probe_after['py'] * np.sqrt(bety)))[0]
qx_probe = (phasex_1 - phasex_0)/(2*np.pi)
qy_probe = (phasey_1 - phasey_0)/(2*np.pi)

#########
# Plots #
#########

import matplotlib.pyplot as plt
plt.close('all')

fig3 = plt.figure(3)
axcoord = fig3.add_subplot(1, 1, 1)
footprint.draw_footprint(xy_norm, axis_object=axcoord, linewidth = 1)
axcoord.set_xlim(right=np.max(xy_norm[:, :, 0]))
axcoord.set_ylim(top=np.max(xy_norm[:, :, 1]))

fig4 = plt.figure(4)
axFP = fig4.add_subplot(1, 1, 1)
footprint.draw_footprint(Qxy_fp, axis_object=axFP, linewidth = 1)
axFP.plot(qx_probe, qy_probe, 'x')
axFP.set_xlim(.15-.07, .15)
axFP.set_ylim(.25-.07, .25)
axFP.set_aspect('equal')
fig4.suptitle(mode)
plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from cpymad.madx import Madx
import json

import xtrack as xt
import xpart as xp
import xobjects as xo

mad_b1 = Madx()

mad_b1.call("lhcb1_seq.madx")
mad_b1.use('lhcb1')
mad_b1.twiss()

line_b1 = xt.Line.from_madx_sequence(sequence=mad_b1.sequence.lhcb1,
    deferred_expressions=True, apply_madx_errors=True, install_apertures=True)
line_b1.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1,
                                    gamma0=mad_b1.sequence.lhcb1.beam.gamma)

with open('line_b1.json', 'w') as fid:
    json.dump(line_b1.to_dict(), fid, cls=xo.JEncoder)

mad_b4 = Madx()
mad_b4.call("lhcb4_seq.madx")
mad_b4.use('lhcb2')
mad_b4.twiss()

line_b4 = xt.Line.from_madx_sequence(sequence=mad_b4.sequence.lhcb2,
    deferred_expressions=True, apply_madx_errors=True, install_apertures=True)

line_b4.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1,
                                    gamma0=mad_b4.sequence.lhcb2.beam.gamma)

with open('line_b4.json', 'w') as fid:
    json.dump(line_b4.to_dict(), fid, cls=xo.JEncoder)






# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time
import numpy as np

from cpymad.madx import Madx

import xtrack as xt

# Import a thick sequence
mad = Madx()
mad.call('../../test_data/clic_dr/sequence.madx')
mad.use('ring')

# Makethin
mad.input(f'''
select, flag=MAKETHIN, SLICE=4, thick=false;
select, flag=MAKETHIN, pattern=wig, slice=1;
MAKETHIN, SEQUENCE=ring, MAKEDIPEDGE=true;
use, sequence=RING;
''')
mad.use('ring')

# Build xtrack line
print('Build xtrack line...')
line = xt.Line.from_madx_sequence(mad.sequence['RING'])
line.particle_ref = xt.Particles(
        mass0=xt.ELECTRON_MASS_EV,
        q0=-1,
        gamma0=mad.sequence.ring.beam.gamma)

# Build line
line.build_tracker()

line.configure_radiation(model='quantum')

record = line.start_internal_logging_for_elements_of_type(
                                                xt.Multipole, capacity=100000)

particles = line.build_particles(x=[0,0,0,0])

line.track(particles, num_turns=10)

import matplotlib.pyplot as plt
hist, bin_edges = np.histogram(record.photon_energy[:record._index.num_recorded], bins=100)
plt.close('all')
plt.loglog(bin_edges[1:], hist/np.diff(bin_edges))
plt.xlabel('Photon energy [eV]')
plt.ylabel('dN/dE [1/eV]')
plt.grid(True)
plt.show()


import numpy as np
from .match import Action
import os
import uuid

from .mad_writer import mad_str_or_value
import xtrack as xt

class MadngVars:

    def __init__(self, mad):
        self.mad = mad

    def __setitem__(self, key, value):
        setattr(self.mad.MADX, key.replace('.', '_'), value)
        #Expressions still to be handled, could use the following:
        # mng.send(
        #     MADX:open_env()
        #     a = 3
        #     b =\ 3 * a
        #     c =\ 4 * a
        #     MADX:close_env()
        #     ''')

def build_madng_model(line, sequence_name='seq'):
    if line.tracker is None:
        line.build_tracker()
    mng = line.to_madng(sequence_name=sequence_name)
    mng._sequence_name = sequence_name
    line.tracker._madng = mng
    line.tracker._madng_vars = MadngVars(mng)
    line.vars.vars_to_update.add(line.tracker._madng_vars)
    return mng

def discard_madng_model(line):
    line.tracker._madng = None
    line.tracker.vars_to_update.remove(line.tracker._madng_vars)
    return

def regen_madng_model(line):
    discard_madng_model(line)
    build_madng_model(line)
    return

def _tw_ng(line, rdts=[], normal_form=True,
           mapdef_twiss=2, mapdef_normal_form=4,
           nslice=3,
           ):

    tw_kwargs = locals()
    del tw_kwargs['line']
    _action = ActionTwissMadng(line, tw_kwargs)

    if not hasattr(line.tracker, '_madng'):
        line.build_madng_model()
    mng = line.tracker._madng

    tw = line.twiss(method='4d', reverse=False)
    tw._action = _action

    tw_columns = ['s', 'beta11', 'beta22', 'alfa11', 'alfa22',
                'x', 'px', 'y', 'py', 't', 'pt',
                'dx', 'dy', 'dpx', 'dpy', 'mu1', 'mu2',
                'beta12', 'beta21', 'alfa12', 'alfa21',
                'wx', 'wy', 'phix', 'phiy', 'dmu1', 'dmu2',
    ]

    columns = tw_columns + rdts
    rdt_cmd = 'local rdts = {"' + '", "'.join(rdts) + '"}'
    send_cmd = f'py:send({{mtbl.{", mtbl.".join(columns)}}})'

    if len(rdts) > 0:
        mng_script = ('''
        local damap in MAD
        '''
        f'local seq = MADX.{mng._sequence_name}'
        '''
        -- list of RDTs
        '''
        + rdt_cmd +
        '''
        -- create phase-space damap at 4th order
        local X0 = damap {nv=6, mo=4}

        -- twiss with RDTs
        local mtbl = twiss {sequence=seq, X0=X0, trkrdt=rdts, info=2, saverdt=true, coupling=true, chrom=true}

        -- send columns to Python
        '''
        + send_cmd)
    else:
        mng_script = ('''
        local damap in MAD
        '''
        f'local seq = MADX.{mng._sequence_name}'
        '''

        -- twiss with RDTs
        local mtbl = twiss {sequence=seq, method=4,'''
        f'mapdef={mapdef_twiss}'
        ''', implicit=true, '''
        f'nslice={nslice}, misalgn=true, coupling=true, chrom=true'
        '''}

        -- send columns to Python
        '''
        + send_cmd)

    mng.send(mng_script)

    out = mng.recv()
    out_dct = {k: v for k, v in zip(columns, out)}

    # Add to table
    assert len(out[0]) == len(tw) + 1
    for nn in tw_columns:
        tw[nn+'_ng'] = np.atleast_1d(np.squeeze(out_dct[nn]))[:-1]
    for nn in rdts:
        tw[nn] = np.atleast_1d(np.squeeze(out_dct[nn]))[:-1]

    temp_x = tw.wx_ng * np.exp(1j*2*np.pi*tw.phix_ng)
    tw['ax_ng'] = np.imag(temp_x)
    tw['bx_ng'] = np.real(temp_x)
    temp_y = tw.wy_ng * np.exp(1j*2*np.pi*tw.phiy_ng)
    tw['ay_ng'] = np.imag(temp_y)
    tw['by_ng'] = np.real(temp_y)
    del tw['phix_ng']
    del tw['phiy_ng']

    if normal_form:
        mng_script_nf = (
            f'local seq = MADX.{mng._sequence_name}'
            '''
            local track in MAD  -- like "from MAD import track"
            local mytrktable, mytrkflow = MAD.track{sequence=seq, method=4,'''
            f'mapdef={mapdef_normal_form}, nslice={nslice}'
            '''}

            local normal in MAD.gphys  -- like "from MAD.gphys import normal"
            local my_norm_for = normal(mytrkflow[1]):analyse('anh') -- anh stands for anharmonicity

            local nf = my_norm_for
            last_nf = my_norm_for
            py:send({
                    nf:q1{1}, -- qx from the normal form (fractional part)
                    nf:q2{1}, -- qy
                    nf:dq1{1}, -- dqx / d delta
                    nf:dq2{1}, -- dqy / d delta
                    nf:dq1{2}, -- d2 qx / d delta2
                    nf:dq2{2}, -- d2 qy / d delta2
                    nf:dq1{3}, -- d3 qx / d delta3
                    nf:dq2{3}, -- d3 qy / d delta3
                    nf:dq1{4}, -- d4 qx / d delta4
                    nf:dq2{4}, -- d4 qy / d delta4
                    nf:dq1{5}, -- d5 qx / d delta5
                    nf:dq2{5}, -- d5 qy / d delta5
                    nf:anhx{1, 0}, -- dqx / d(2 jx)
                    nf:anhy{0, 1}, -- dqy / d(2 jy)
                    nf:anhx{0, 1}, -- dqx / d(2 jy)
                    nf:anhy{1, 0}, -- dqy / d(2 jx)
                    })
        ''')
        mng.send(mng_script_nf)
        out_nf = mng.recv()

        dct_nf = dict(
            q1 =   out_nf[0],
            q2 =   out_nf[1],
            dq1 =  out_nf[2],
            dq2 =  out_nf[3],
            d2q1 = out_nf[4],
            d2q2 = out_nf[5],
            d3q1 = out_nf[6],
            d3q2 = out_nf[7],
            d4q1 = out_nf[8],
            d4q2 = out_nf[9],
            d5q1 = out_nf[10],
            d5q2 = out_nf[11],
            dqxdjx = out_nf[6]*2.,
            dqydjy = out_nf[7]*2.,
            dqxdjy = out_nf[8]*2.,
            dqydjx = out_nf[9]*2.,
        )
        for nn in dct_nf:
            tw[nn+'_nf_ng'] = dct_nf[nn]

    return tw

class ActionTwissMadng(Action):
    def __init__(self, line, tw_kwargs):
        self.line = line
        self.tw_kwargs = tw_kwargs

    def run(self):
        return self.line.madng_twiss(**self.tw_kwargs)

def line_to_madng(line, sequence_name='seq', temp_fname=None, keep_files=False):

    try:
        _ge = xt.elements._get_expr
        if temp_fname is None:
            temp_fname = 'temp_madng_' + str(uuid.uuid4())

        madx_seq = line.to_madx_sequence(sequence_name=sequence_name)
        with open(f'{temp_fname}.madx', 'w') as fid:
            fid.write(madx_seq)

        from pymadng import MAD
        mng = MAD()
        mng.MADX.load(f'"{temp_fname}.madx"', f'"{temp_fname}"')
        mng._init_madx_data = madx_seq

        mng[sequence_name] = mng.MADX[sequence_name] # this ensures that the file has been read
        mng[sequence_name].beam = mng.beam(particle="'custom'",
                        mass=line.particle_ref.mass0 * 1e9,
                        charge=line.particle_ref.q0,
                        betgam=line.particle_ref.beta0[0] * line.particle_ref.gamma0[0])

        # Patch shifts (MAD-NG ignores dx, dy from MAD-X, need to set them through misalign)
        commands = []
        commands.append('MADX:open_env()')
        for nn in line.element_names:
            if not hasattr(line[nn], 'shift_x'):
                continue
            nn_ng = nn.replace('.', '_')
            dx = mad_str_or_value(_ge(line.ref[nn].shift_x))
            dy = mad_str_or_value(_ge(line.ref[nn].shift_y))
            if dx == 0 and dy == 0:
                continue
            commands.append(
                f'{nn_ng}.dx = 0\n'
                f'{nn_ng}.dy = 0\n'
                f'{nn_ng}.misalign'
                ' =\\ {'
                f'dx={dx},'
                f'dy={dy}'
                '}'
            )
        commands.append('MADX:close_env()')
        mng.send('\n'.join(commands))

    finally:
        if not keep_files:
            for nn in [temp_fname + '.madx', temp_fname + '.mad']:
                if os.path.isfile(nn):
                    os.remove(nn)

    # mng[sequence_name].beam = mng.beam(particle="'proton'", energy=7000)
    return mng